import beeline
from beeline.propagation import Request
from beeline.middleware.wsgi import WSGIRequest


class HoneyWSGIMiddleware(object):

    def __init__(self, app):
        self.app = app

    def __call__(self, environ, start_response):
        wr = WSGIRequest("bottle", environ)

        root_span = beeline.propagate_and_start_trace(wr.request_context(), wr)

        def _start_response(status, headers, *args):
            beeline.add_context_field("response.status_code", status)
            beeline.finish_trace(root_span)

            return start_response(status, headers, *args)

        return self.app(environ, _start_response)



import beeline
from beeline.propagation import PropagationContext
import re

# Cribbed from OpenTelemetry python implementation.
_TRACEPARENT_HEADER_FORMAT = (
    "^[ \t]*([0-9a-f]{2})-([0-9a-f]{32})-([0-9a-f]{16})-([0-9a-f]{2})"
    + "(-.*)?[ \t]*$"
)
_TRACEPARENT_HEADER_FORMAT_RE = re.compile(_TRACEPARENT_HEADER_FORMAT)
_EMPTY_TRACE_ID = "0" * 32
_EMPTY_PARENT_ID = "0" * 16


def http_trace_parser_hook(request):
    '''
    Retrieves the w3c propagation context out of the request.
    request must implement the beeline.propagation.Request abstract base class
    '''
    traceparent_header = request.header('traceparent')
    if not traceparent_header:
        return None
    tracestate_header = request.header('tracestate')

    trace_id = None
    parent_id = None
    try:
        trace_id, parent_id, trace_flags = unmarshal_traceparent(
            traceparent_header)
        tracestate = unmarshal_tracestate(tracestate_header)
        trace_fields = {}
        trace_fields['traceflags'] = trace_flags
        trace_fields['tracestate'] = tracestate
        return PropagationContext(trace_id, parent_id, trace_fields)
    except Exception as e:
        beeline.internal.log(
            'error attempting to extract w3c trace: %s', beeline.internal.stringify_exception(e))
    return None


def http_trace_propagation_hook(propagation_context):
    '''
    Given a propagation context, returns a dictionary of key value pairs that should be
    added to outbound requests (usually HTTP headers)
    '''
    if not propagation_context:
        return None

    traceparent_header = marshal_traceparent(propagation_context)
    if not traceparent_header:
        return {}

    headers = {}
    headers["traceparent"] = traceparent_header

    tracestate_header = marshal_tracestate(propagation_context)

    if tracestate_header:
        headers['tracestate'] = tracestate_header
    return headers


def marshal_traceparent(propagation_context):
    '''
    Given a propagation context, returns the contents of a trace header to be
    injected by middleware.
    '''
    if not propagation_context:
        return None

    trace_flags = propagation_context.trace_fields.get('traceflags')
    if not trace_flags:
        trace_flags = "01"

    traceparent_header = "00-{}-{}-{}".format(
        propagation_context.trace_id,
        propagation_context.parent_id,
        trace_flags,
    )

    return traceparent_header


def marshal_tracestate(propagation_context):
    '''
    '''
    if not propagation_context:
        return None

    tracestate_header = propagation_context.trace_fields.get('tracestate')
    return tracestate_header


def unmarshal_traceparent(header):
    match = re.search(_TRACEPARENT_HEADER_FORMAT_RE, header)
    if not match:
        # Raise exception?
        return None
    version = match.group(1)
    trace_id = match.group(2)
    parent_id = match.group(3)
    trace_flags = match.group(4)

    if trace_id == _EMPTY_TRACE_ID or parent_id == _EMPTY_PARENT_ID:
        return None

    if version == "00":
        if match.group(5):
            return None
    if version == "ff":
        return None

    return trace_id, parent_id, trace_flags


def unmarshal_tracestate(header):
    # We treat the tracestate header as an opaque blob, and don't parse it at all.
    return header



import unittest
from mock import Mock, patch, ANY

from beeline.middleware import awslambda

header_value = '1;trace_id=bloop,parent_id=scoop,context=e30K'


class TestLambdaRequest(unittest.TestCase):
    def test_get_header_from_headers(self):
        '''
        Test that if headers is set, we have case-insensitive match.
        '''
        event = {
            'headers': {
                # case shouldn't matter
                'X-HoNEyComb-TrACE': header_value,
            },
        }

        lr = awslambda.LambdaRequest(event)
        self.assertIsNotNone(lr)
        self.assertEqual(lr.header('X-Honeycomb-Trace'), header_value)

    def test_handle_no_headers(self):
        ''' ensure that we handle events with no header key '''
        event = {
            'foo': 1,
        }

        lr = awslambda.LambdaRequest(event)
        self.assertIsNotNone(lr)
        self.assertIsNone(lr.header('X-Honeycomb-Trace'))

    def test_handle_sns_none(self):
        ''' ensure that we handle SNS events with no honeycomb key '''
        event = {
            "Records":
            [
                {
                    "EventSource": "aws:sns",
                    "Sns": {
                        "Message": "Hello from SNS!",
                        "MessageAttributes": {}
                    }
                }
            ],
        }

        lr = awslambda.LambdaRequest(event)
        self.assertIsNotNone(lr)
        self.assertIsNone(lr.header('X-Honeycomb-Trace'))

    def test_handle_sns_attribute(self):
        ''' ensure that we extract SNS data from message attributes'''
        event = {
            "Records":
            [
                {
                    "EventSource": "aws:sns",
                    "Sns": {
                        "Message": "Hello from SNS!",
                        "MessageAttributes": {
                            'X-HoNEyComb-TrACE': {
                                "Type": "String",
                                "Value": header_value,
                            }
                        }
                    }
                }
            ],
        }

        lr = awslambda.LambdaRequest(event)
        self.assertIsNotNone(lr)
        self.assertEqual(lr.header('X-Honeycomb-Trace'), header_value)

    def test_handle_sqs_none(self):
        ''' ensure that we handle SQS events with no honeycomb key '''
        event = {
            "Records":
            [
                {
                    "body": "Hello from SQS!",
                    "messageAttributes": {},
                    "eventSource": "aws:sqs",
                },
            ],
        }

        lr = awslambda.LambdaRequest(event)
        self.assertIsNotNone(lr)
        self.assertIsNone(lr.header('X-Honeycomb-Trace'))

    def test_handle_sqs_attributes(self):
        ''' ensure that we extract SQS data from message attributes'''
        event = {
            "Records":
            [
                {
                    "body": "Hello from SQS!",
                    "messageAttributes": {
                        'X-HoNEyComb-TrACE': {
                            "Type": "String",
                            "stringValue": header_value,
                        },
                        'foo': {
                            "Type": "String",
                            "stringValue": "bar",
                        },
                    },
                    "eventSource": "aws:sqs",
                },
            ],
        }

        lr = awslambda.LambdaRequest(event)
        self.assertIsNotNone(lr)
        self.assertEqual(lr.header('X-Honeycomb-Trace'), header_value)

    def test_message_batch_is_ignored(self):
        ''' ensure that we don't process batches'''
        event = {
            "Records":
            [
                {
                    "body": "Hello from SQS!",
                    "messageAttributes": {
                        'X-HoNEyComb-TrACE': {
                            "Type": "String",
                            "stringValue": "1;trace_id=beep,parent_id=moop,context=e29K",
                        },
                        'foo': {
                            "Type": "String",
                            "stringValue": "bar",
                        },
                    },
                    "eventSource": "aws:sqs",
                },
                {
                    "body": "Another hello from SQS!",
                    "messageAttributes": {
                        'X-HoNEyComb-TrACE': {
                            "Type": "String",
                            "stringValue": "1;trace_id=bloop,parent_id=scoop,context=e30K",
                        },
                        'foo': {
                            "Type": "String",
                            "stringValue": "baz",
                        },
                    },
                    "eventSource": "aws:sqs",
                },
            ],
        }

        lr = awslambda.LambdaRequest(event)
        self.assertIsNotNone(lr)
        self.assertIsNone(lr.header('X-Honeycomb-Trace'))


class TestLambdaWrapper(unittest.TestCase):
    def test_wrapper_works_no_init(self):
        ''' ensure that the wrapper doesn't break anything if used before
        beeline.init is called
        '''
        with patch('beeline.get_beeline') as p:
            p.return_value = None

            @awslambda.beeline_wrapper
            def foo(event, context):
                return 1

            self.assertEqual(foo(None, None), 1)

            @awslambda.beeline_wrapper()
            def bar(event, context):
                return 1

            self.assertEqual(bar(None, None), 1)

    def test_basic_instrumentation(self):
        ''' ensure basic event fields get instrumented '''
        with patch('beeline.propagate_and_start_trace') as m_propagate, \
                patch('beeline.add_context_field') as m_add_context_field, \
                patch('beeline.middleware.awslambda.beeline._GBL'), \
                patch('beeline.middleware.awslambda.COLD_START') as m_cold_start:
            m_event = Mock()
            m_context = Mock(function_name='fn', function_version="1.1.1",
                             aws_request_id='12345')

            @awslambda.beeline_wrapper
            def handler(event, context):
                return 1

            self.assertEqual(handler(m_event, m_context), 1)
            m_propagate.assert_called_once_with({
                'app.function_name': 'fn',
                'app.function_version': '1.1.1',
                'app.request_id': '12345',
                'app.event': ANY,  # 'app.event' is included by default
                'meta.cold_start': ANY,
                'name': 'handler'}, ANY)
            m_add_context_field.assert_called_once_with('app.response', 1)

    def test_handle_exceptions(self):
        ''' ensure instrumentation occurs when the handler raises an exception '''
        with patch('beeline.propagate_and_start_trace') as m_propagate, \
                patch('beeline.add_context_field') as m_add_context_field, \
                patch('beeline.add_context') as m_add_context, \
                patch('beeline.middleware.awslambda.beeline._GBL'), \
                patch('beeline.middleware.awslambda.COLD_START') as m_cold_start:
            m_event = Mock()
            m_context = Mock(function_name='fn', function_version="1.1.1",
                             aws_request_id='12345')

            @awslambda.beeline_wrapper
            def handler(event, context):
                raise ValueError('something went wrong')

            with self.assertRaises(ValueError):
                handler(m_event, m_context)

            m_propagate.assert_called_once_with({
                'app.function_name': 'fn',
                'app.function_version': '1.1.1',
                'app.request_id': '12345',
                'app.event': ANY,  # 'app.event' is included by default
                'meta.cold_start': ANY,
                'name': 'handler'}, ANY)
            m_add_context.assert_called_once_with({
                'app.exception_type': ANY,  # representation changes between 2.7 and 3.x
                'app.exception_string': 'something went wrong',
                'app.exception_stacktrace': ANY
            })

    def test_can_omit_input(self):
        ''' ensure input event field can be omitted '''
        with patch('beeline.propagate_and_start_trace') as m_propagate, \
                patch('beeline.add_context_field') as m_add_context_field, \
                patch('beeline.middleware.awslambda.beeline._GBL'), \
                patch('beeline.middleware.awslambda.COLD_START') as m_cold_start:
            m_event = Mock()
            m_context = Mock(function_name='fn', function_version="1.1.1",
                             aws_request_id='12345')

            @awslambda.beeline_wrapper(record_input=False)
            def handler(event, context):
                return 1

            self.assertEqual(handler(m_event, m_context), 1)
            m_propagate.assert_called_once_with({
                'app.function_name': 'fn',
                'app.function_version': '1.1.1',
                'app.request_id': '12345',
                'meta.cold_start': ANY,
                'name': 'handler'}, ANY)  # note the lack of an 'app.event' field
            m_add_context_field.assert_called_once_with('app.response', 1)

    def test_can_omit_output(self):
        ''' ensure output event fields can be omitted '''
        with patch('beeline.propagate_and_start_trace') as m_propagate, \
                patch('beeline.add_context_field') as m_add_context_field, \
                patch('beeline.middleware.awslambda.beeline._GBL'), \
                patch('beeline.middleware.awslambda.COLD_START') as m_cold_start:
            m_event = Mock()
            m_context = Mock(function_name='fn', function_version="1.1.1",
                             aws_request_id='12345')

            @awslambda.beeline_wrapper(record_output=False)
            def handler(event, context):
                return 1

            self.assertEqual(handler(m_event, m_context), 1)
            m_propagate.assert_called_once_with({
                'app.function_name': 'fn',
                'app.function_version': '1.1.1',
                'app.request_id': '12345',
                'app.event': ANY,  # 'app.event' is included by default
                'meta.cold_start': ANY,
                'name': 'handler'}, ANY)
            m_add_context_field.not_called_with('app.response', 1)



from wrapt import wrap_function_wrapper
import beeline
import beeline.propagation
import urllib.request


def _urllibopen(_urlopen, instance, args, kwargs):
    # urlopen accepts either a string URL or a Request object as its first arg
    # It's easier to process the info contained in the request and modify it
    # by converting the URL string into a Request
    if type(args[0]) != urllib.request.Request:
        args = (urllib.request.Request(args[0]),) + tuple(args[1:])

    span = beeline.start_span(context={"meta.type": "http_client"})

    b = beeline.get_beeline()
    if b and b.http_trace_propagation_hook is not None:
        new_headers = beeline.http_trace_propagation_hook()
        if new_headers:
            # Merge the new headers into the existing headers for the outbound request
            b.log(
                "urllib lib - adding trace context to outbound request: %s", new_headers)
            args[0].headers.update(new_headers)

    try:
        resp = None
        beeline.add_context({
            "name": "urllib_%s" % args[0].get_method(),
            "request.method": args[0].get_method(),
            "request.uri": args[0].full_url
        })
        resp = _urlopen(*args, **kwargs)
        return resp
    except Exception as e:
        beeline.add_context({
            "request.error_type": str(type(e)),
            "request.error": beeline.internal.stringify_exception(e),
        })
        raise
    finally:
        if resp:
            beeline.add_context_field("response.status_code", resp.status)
            content_type = resp.getheader('content-type')
            if content_type:
                beeline.add_context_field(
                    "response.content_type", content_type)
            content_length = resp.getheader('content-length')
            if content_length:
                beeline.add_context_field(
                    "response.content_length", content_length)

        beeline.finish_span(span)


# Note that this only patches urllib.request.urlopen, not
# http.client.HTTPConnection.  The latter is a lot more of a pain to figure
# out what, exactly, the lifetime of the span ought to be -- but most people
# who plan to block and do nothing else use urlopen, anyway.
wrap_function_wrapper('urllib.request', 'urlopen', _urllibopen)



import beeline
from beeline.propagation import PropagationContext
import base64
import json
from six.moves.urllib.parse import quote, unquote


def http_trace_parser_hook(request):
    '''
    Retrieves the honeycomb propagation context out of the request.
    request must implement the beeline.propagation.Request abstract base class
    '''
    trace_header = request.header('X-Honeycomb-Trace')
    if trace_header:
        try:
            trace_id, parent_id, context, dataset = unmarshal_propagation_context_with_dataset(
                trace_header)
            return PropagationContext(trace_id, parent_id, context, dataset)
        except Exception as e:
            beeline.internal.log(
                'error attempting to extract trace context: %s', beeline.internal.stringify_exception(e))
    return None


def http_trace_propagation_hook(propagation_context):
    '''
    Given a propagation context, returns a dictionary of key value pairs that should be
    added to outbound requests (usually HTTP headers)
    '''
    if not propagation_context:
        return None

    return {"X-Honeycomb-Trace": marshal_propagation_context(propagation_context)}


def marshal_propagation_context(propagation_context):
    '''
    Given a propagation context, returns the contents of a trace header to be
    injected by middleware.
    '''
    if not propagation_context:
        return None

    # FIXME: Since ALL trace fields are in propagation_context, we may want to strip
    # some automatically added trace fields that we DON'T want to propagate - e.g. request.*
    version = 1
    trace_fields = base64.b64encode(json.dumps(
        propagation_context.trace_fields).encode()).decode()

    components = ["trace_id={}".format(propagation_context.trace_id),
                  "parent_id={}".format(propagation_context.parent_id),
                  "context={}".format(trace_fields)]

    if beeline.propagation.propagate_dataset and propagation_context.dataset:
        components.insert(0, "dataset={}".format(quote(propagation_context.dataset)))

    trace_header = "{};{}".format(version, ",".join(components))

    return trace_header


def unmarshal_propagation_context(trace_header):
    """
    Deprecated: Use beeline.propagation.honeycomb.unmarshal_propagation_context_with_dataset instead
    """
    trace_id, parent_id, context, _dataset = unmarshal_propagation_context_with_dataset(trace_header)

    return trace_id, parent_id, context


def unmarshal_propagation_context_with_dataset(trace_header):
    '''
    Given the body of the `X-Honeycomb-Trace` header, returns the trace_id,
    parent_id, "context" and dataset.
    '''
    # the first value is the trace payload version
    # at this time there is only one version, but we should warn
    # if another version comes through
    version, data = trace_header.split(';', 1)
    if version != "1":
        beeline.internal.log(
            'warning: trace_header version %s is unsupported', version)
        return None, None, None, None

    kv_pairs = data.split(',')

    trace_id, parent_id, context, dataset = None, None, None, None
    # Some beelines send "dataset" but we do not handle that yet
    for pair in kv_pairs:
        k, v = pair.split('=', 1)
        if k == 'trace_id':
            trace_id = v
        elif k == 'parent_id':
            parent_id = v
        elif k == 'context':
            context = json.loads(base64.b64decode(v.encode()).decode())
        elif k == 'dataset' and beeline.propagation.propagate_dataset:
            dataset = unquote(v)

    # context should be a dict
    if context is None:
        context = {}

    return trace_id, parent_id, context, dataset



import unittest
from mock import Mock, call, patch

import django
from django.http import HttpResponse, StreamingHttpResponse
from django.test.client import Client
from django.conf.urls import url

from beeline.middleware.django import HoneyMiddlewareBase


class SimpleWSGITest(unittest.TestCase):
    def setUp(self):
        self.addCleanup(patch.stopall)
        self.m_gbl = patch('beeline.middleware.django.beeline').start()

    def test_call_middleware(self):
        ''' Just call the middleware and ensure that the code runs '''
        mock_req = Mock()
        mock_resp = Mock()
        mock_resp.return_value.streaming = False
        mock_trace = Mock()
        self.m_gbl.propagate_and_start_trace.return_value = mock_trace

        mw = HoneyMiddlewareBase(mock_resp)
        resp = mw(mock_req)
        self.m_gbl.propagate_and_start_trace.assert_called_once()

        mock_resp.assert_called_once_with(mock_req)

        self.m_gbl.finish_trace.assert_called_once_with(mock_trace)
        self.assertEqual(resp, mock_resp.return_value)


@unittest.skipIf(django.VERSION < (2, 2), "Routes are only supported on Django 2.2 and higher")
class FullViewTestCase(unittest.TestCase):
    def setUp(self):
        self.addCleanup(patch.stopall)
        self.m_gbl = patch('beeline.middleware.django.beeline').start()

        # Unfortunately we need to import these quite late, because if we use a
        # top-level import, the test discovery procedure checks if `settings`
        # is a subclass of `TestCase`, causing the settings to be initialized,
        # which isn't possible.
        from django.conf import settings  # pylint: disable=bad-option-value,import-outside-toplevel
        from django.utils.functional import empty  # pylint: disable=bad-option-value,import-outside-toplevel
        assert not settings.configured
        # On shutdown:
        self.addCleanup(lambda: setattr(settings, "_wrapped", empty))
        settings.configure(
            MIDDLEWARE=['beeline.middleware.django.HoneyMiddlewareHttp'],
            ALLOWED_HOSTS=['testserver'],
            ROOT_URLCONF=(
                url("^hello/(?P<greetee>[^/]+)/$", self._view, name="greet"),
                url("^stream/hello/(?P<greetee>[^/]+)/$", self._streaming_view, name="stream_greet"),
            ),
        )

    def _view(self, request, *args, **kwargs):
        return HttpResponse(kwargs["greetee"], status=200)

    def _streaming_view(self, request, *args, **kwargs):
        def stream():
            yield b"hello "
            yield kwargs["greetee"]
        return StreamingHttpResponse(stream(), status=200)

    def test_middleware(self):
        mock_trace = Mock()
        self.m_gbl.propagate_and_start_trace.return_value = mock_trace

        response = Client().get('/hello/world/')
        self.assertEqual(response.content, b"world")

        self.m_gbl.add_context_field.assert_has_calls([
            call("django.view_func", "_view"),
            call("request.route", "^hello/(?P<greetee>[^/]+)/$"),
        ])
        self.m_gbl.finish_trace.assert_called_once_with(mock_trace)

    def test_streaming_middleware(self):
        mock_trace = Mock()
        self.m_gbl.propagate_and_start_trace.return_value = mock_trace

        response = Client().get('/stream/hello/world/')
        self.m_gbl.finish_trace.assert_not_called()
        content = b"".join(response.streaming_content)
        self.m_gbl.finish_trace.assert_called_once_with(mock_trace)
        self.assertEqual(content, b"hello world")



''' module beeline '''
import functools
import logging
import os
import socket
from contextlib import contextmanager

from libhoney import Client
from beeline.trace import SynchronousTracer
from beeline.version import VERSION
from beeline import internal
import beeline.propagation.default
import beeline.propagation
import sys
# pyflakes
assert internal

USER_AGENT_ADDITION = "beeline-python/%s" % VERSION

# This is the global beeline created by init
_GBL = None
# This is the PID that initialized the beeline.
_INITPID = None

try:
    import asyncio
    try:
        asyncio.get_running_loop()  # pylint: disable=no-member
    except RuntimeError:
        pass

    from beeline.aiotrace import AsyncioTracer, traced_impl, untraced
    assert untraced

    def in_async_code():
        """Return whether we are running inside an asynchronous task.

        We use this information to determine which tracer
        implementation to use.

        """
        try:
            asyncio.get_running_loop()  # pylint: disable=no-member
            return True
        except RuntimeError:
            return False

except (ImportError, AttributeError):
    # Use these non-async versions if we don't have asyncio.
    from beeline.trace import traced_impl

    def in_async_code():
        return False


class Beeline(object):
    def __init__(self,
                 writekey='', dataset='', service_name='',
                 tracer=None, sample_rate=1, api_host='https://api.honeycomb.io',
                 max_concurrent_batches=10, max_batch_size=100, send_frequency=0.25,
                 block_on_send=False, block_on_response=False,
                 transmission_impl=None, sampler_hook=None, presend_hook=None,
                 http_trace_parser_hook=beeline.propagation.default.http_trace_parser_hook,
                 http_trace_propagation_hook=beeline.propagation.default.http_trace_propagation_hook,
                 debug=False):

        self.client = None
        self.tracer_impl = None
        self.presend_hook = None
        self.sampler_hook = None
        self.http_trace_parser_hook = None
        self.http_trace_propagation_hook = None

        self.debug = debug
        if debug:
            self._init_logger()

        def IsClassicKey(writekey):
            return len(writekey) == 32

        # allow setting some values from the environment
        if not writekey:
            writekey = os.environ.get('HONEYCOMB_WRITEKEY', '')
        # also check API_KEY just in case
        if not writekey:
            writekey = os.environ.get('HONEYCOMB_API_KEY', '')
        if not writekey:
            logging.error(
                'writekey not set! set the writekey if you want to send data to honeycomb'
            )

        if IsClassicKey(writekey):
            if not dataset:
                dataset = os.environ.get('HONEYCOMB_DATASET', '')
            if not dataset:
                logging.error(
                    'dataset not set! set a value for dataset if you want to send data to honeycomb'
                )
        else:
            if dataset:
                logging.error(
                    'dataset will be ignored in favor of service name'
                )

        default_service_name = "unknown_service"
        process_name = os.environ.get('PROCESS_EXECUTABLE_NAME', '')
        if not service_name:
            service_name = os.environ.get('HONEYCOMB_SERVICE')
        # also check SERVICE_NAME just in case
        if not service_name:
            service_name = os.environ.get('SERVICE_NAME', '')
        # no service name, set default
        if not service_name:
            service_name = default_service_name
            if process_name:
                service_name += ":" + process_name
            else:
                service_name += ":python"
            logging.error(
                'service name not set! service name will be ' + service_name
            )
            if not IsClassicKey(writekey):
                logging.error(
                    'data will be sent to unknown_service'
                )

        if not IsClassicKey(writekey):
            # set dataset based on service name
            dataset = service_name
            if dataset.strip() != dataset:
                # whitespace detected. trim whitespace, warn on diff
                logging.error(
                    'service name has unexpected spaces'
                )
                dataset = service_name.strip()

            # set default, truncate to unknown_service if needed
            if dataset == "" or dataset.startswith("unknown_service"):
                dataset = "unknown_service"

        self.client = Client(
            writekey=writekey, dataset=dataset, sample_rate=sample_rate,
            api_host=api_host, max_concurrent_batches=max_concurrent_batches,
            max_batch_size=max_batch_size, send_frequency=send_frequency,
            block_on_send=block_on_send, block_on_response=block_on_response,
            transmission_impl=transmission_impl,
            user_agent_addition=USER_AGENT_ADDITION,
            debug=debug,
        )

        if IsClassicKey(writekey):
            beeline.propagation.propagate_dataset = True
        else:
            beeline.propagation.propagate_dataset = False

        self.log('initialized honeycomb client: writekey=%s dataset=%s service_name=%s',
                 writekey, dataset, service_name)

        self.client.add_field('service_name', service_name)
        self.client.add_field('service.name', service_name)
        self.client.add_field('meta.beeline_version', VERSION)
        self.client.add_field('meta.local_hostname', socket.gethostname())

        if in_async_code():
            self.tracer_impl = AsyncioTracer(self.client)
        else:
            self.tracer_impl = SynchronousTracer(self.client)
        self.tracer_impl.register_hooks(
            presend=presend_hook,
            sampler=sampler_hook,
            http_trace_parser=http_trace_parser_hook,
            http_trace_propagation=http_trace_propagation_hook)
        self.sampler_hook = sampler_hook
        self.presend_hook = presend_hook
        self.http_trace_parser_hook = http_trace_parser_hook
        self.http_trace_propagation_hook = http_trace_propagation_hook

    def send_now(self, data):
        ''' DEPRECATED - to be removed in a future release

        Create an event and enqueue it immediately. Does not work with
        `beeline.add_field` - this is equivalent to calling `libhoney.send_now`
        '''
        ev = self.client.new_event()

        if data:
            ev.add(data)
        self._run_hooks_and_send(ev)

    def add_field(self, name, value):
        ''' Add a field to the currently active span.

        `beeline.add_field("my field", "my value")`

        If a field is being attributed to the wrong span/event,
        make sure that `new_event` and `close_event` calls are matched.
        '''
        # fetch the current event from our tracer
        span = self.tracer_impl.get_active_span()
        # if there are no spans, this is a noop
        if span is None:
            return
        span.add_context_field(name, value)

    def add(self, data):
        '''Similar to add_field(), but allows you to add a number of name:value pairs
        to the currently active event at the same time.

        `beeline.add({ "first_field": "a", "second_field": "b"})`
        '''
        # fetch the current event from the tracer
        span = self.tracer_impl.get_active_span()
        # if there are no spans, this is a noop
        if span is None:
            return

        span.add_context(data)

    def tracer(self, name, trace_id=None, parent_id=None):
        return self.tracer_impl(name=name, trace_id=trace_id, parent_id=parent_id)

    def new_event(self, data=None, trace_name=''):
        ''' DEPRECATED: Helper method that wraps `start_trace` and
        `start_span`. It is better to use these methods as it provides
        better control and context around how traces are implemented in your
        app.

        Creates a new span, populating it with the given data if
        supplied. If no trace is running, a new trace will be started,
        otherwise the event will be added as a span of the existing trace.

        To send the event, call `beeline.send_event()`. There should be a
        `send_event()` for each call to `new_event()`, or tracing and
        `add` and `add_field` will not work correctly.

        If trace_name is specified, will set the "name" field of the current span,
        which is used in the trace visualizer.
        '''
        if trace_name:
            data['name'] = trace_name

        if self.tracer_impl.get_active_trace_id():
            self.tracer_impl.start_span(context=data)
        else:
            self.tracer_impl.start_trace(context=data)

    def send_event(self):
        ''' DEPRECATED: Sends the currently active event (current span),
        if it exists.

        There must be one call to `send_event` for each call to `new_event`.
        '''

        span = self.tracer_impl.get_active_span()
        if span:
            if span.is_root():
                self.tracer_impl.finish_trace(span)
                return
            self.tracer_impl.finish_span(span)

    def send_all(self):
        ''' send all spans in the trace stack, regardless of their
        state
        '''

        span = self.tracer_impl.get_active_span()
        while span:
            if span.is_root():
                self.tracer_impl.finish_trace(span)
                return
            self.tracer_impl.finish_span(span)
            span = self.tracer_impl.get_active_span()

    def traced(self, name, trace_id=None, parent_id=None):
        return traced_impl(tracer_fn=self.tracer, name=name, trace_id=trace_id, parent_id=parent_id)

    def traced_thread(self, fn):
        trace_copy = self.tracer_impl._trace.copy()

        @functools.wraps(fn)
        def wrapped(*args, **kwargs):
            self.tracer_impl._trace = trace_copy
            return fn(*args, **kwargs)

        return wrapped

    def _run_hooks_and_send(self, ev):
        ''' internal - run any defined hooks on the event and send '''
        presampled = False
        if self.sampler_hook:
            self.log("executing sampler hook on event ev = %s", ev.fields())
            keep, new_rate = self.sampler_hook(ev.fields())
            if not keep:
                self.log(
                    "skipping event due to sampler hook sampling ev = %s", ev.fields())
                return
            ev.sample_rate = new_rate
            presampled = True

        if self.presend_hook:
            self.log("executing presend hook on event ev = %s", ev.fields())
            self.presend_hook(ev.fields())

        if presampled:
            self.log("enqueuing presampled event ev = %s", ev.fields())
            ev.send_presampled()
        else:
            self.log("enqueuing event ev = %s", ev.fields())
            ev.send()

    def _init_logger(self):
        self._logger = logging.getLogger('honeycomb-beeline')
        self._logger.setLevel(logging.DEBUG)
        ch = logging.StreamHandler()
        ch.setLevel(logging.DEBUG)
        formatter = logging.Formatter(
            '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        ch.setFormatter(formatter)
        self._logger.addHandler(ch)

    def log(self, msg, *args, **kwargs):
        if self.debug:
            self._logger.debug(msg, *args, **kwargs)

    def get_responses_queue(self):
        return self.client.responses()

    def close(self):
        if self.client:
            self.client.close()


def init(writekey='', dataset='', service_name='', tracer=None,
         sample_rate=1, api_host='https://api.honeycomb.io', transmission_impl=None,
         sampler_hook=None, presend_hook=None, debug=False, *args, **kwargs):
    ''' initialize the honeycomb beeline. This will initialize a libhoney
    client local to this module, and a tracer to track traces and spans.

    Args:
    - `writekey`: the authorization key for your team on Honeycomb. Find your team
            write key at [https://ui.honeycomb.io/account](https://ui.honeycomb.io/account)
    - `dataset`: the name of the default dataset to which to write
    - `sample_rate`: the default sample rate. 1 / `sample_rate` events will be sent.
    - `transmission_impl`: if set, override the default transmission implementation
            (for example, TornadoTransmission)
    - `sampler_hook`: accepts a function to be called just before each event is sent.
            The function should accept a dictionary of event fields, and return a tuple
            of type (bool, int). The first item indicates whether or not the event
            should be sent, and the second indicates the updated sample rate to use.
    - `presend_hook`: accepts a function to be called just before each event is sent.
            The functon should accept a dictionary of event fields, and can be used
            to add new fields, modify/scrub existing fields, or drop fields. This
            function is called after sampler_hook, if sampler_hook is set.

    If in doubt, just set `writekey` and `dataset` and move on!
    '''
    global _GBL
    global _INITPID
    pid = os.getpid()
    if _GBL:
        if pid == _INITPID:
            _GBL.log("beeline already initialized! skipping initialization")
            return
        _GBL.log("beeline already initialized, but process ID has changed (was {}, now {}). Reinitializing.".format(
            _INITPID, pid))
        _GBL.close()

    _GBL = Beeline(
        writekey=writekey, dataset=dataset, sample_rate=sample_rate,
        api_host=api_host, transmission_impl=transmission_impl,
        debug=debug, presend_hook=presend_hook, sampler_hook=sampler_hook,
        service_name=service_name,
        # since we've simplified the init function signature a bit,
        # pass on other args for backwards compatibility
        *args, **kwargs
    )
    # Store the PID that initialized the beeline globally. If the beeline was initialized in another
    # process that was later forked, we can use this to detect it and reinitialize the client (and the transmission
    # thread).
    _INITPID = pid


def send_now(data):
    ''' Create an event and enqueue it immediately. Does not work with
    `beeline.add_field` - this is equivalent to calling `libhoney.send_now`

    Args:
    - `data`: dictionary of field names (strings) to field values to include
              in the event
    '''
    # no-op if we're not initialized
    bl = get_beeline()

    if bl:
        bl.send_now(data)


def add_field(name, value):
    ''' DEPRECATED: use `add_context_field`

    Args:
    - `data`: dictionary of field names (strings) to field values to add
    '''
    if _GBL:
        _GBL.add_field(name, value)


def add(data):
    '''DEPRECATED: use `add_context`

    Args:
    - `data`: dictionary of field names (strings) to field values to add
    '''
    bl = get_beeline()

    if bl:
        bl.add(data)


def add_context(data):
    '''Similar to add_context_field(), but allows you to add a number of name:value pairs
    to the currently active event at the same time.

    `beeline.add_context({ "first_field": "a", "second_field": "b"})`

    Args:
    - `data`: dictionary of field names (strings) to field values to add
    '''
    bl = get_beeline()

    if bl:
        bl.tracer_impl.add_context(data=data)


def add_context_field(name, value):
    ''' Add a field to the currently active span. For example, if you are
    using django and wish to add additional context to the current request
    before it is sent to Honeycomb:

    `beeline.add_context_field("my field", "my value")`

    Args:
    - `name`: Name of field to add
    - `value`: Value of new field
    '''
    bl = get_beeline()

    if bl:
        bl.tracer_impl.add_context_field(name=name, value=value)


def remove_context_field(name):
    ''' Remove a single field from the current span.

    ```
    beeline.add_context({ "first_field": "a", "second_field": "b"})
    beeline.remove_context_field("second_field")

    Args:
    - `name`: Name of field to remove
    ```
     '''
    bl = get_beeline()

    if bl:
        bl.tracer_impl.remove_context_field(name=name)


def add_rollup_field(name, value):
    ''' AddRollupField adds a key/value pair to the current span. If it is called repeatedly
    on the same span, the values will be summed together.  Additionally, this
    field will be summed across all spans and added to the trace as a total. It
    is especially useful for doing things like adding the duration spent talking
    to a specific external service - eg database time. The root span will then
    get a field that represents the total time spent talking to the database from
    all of the spans that are part of the trace.

    Args:
    - `name`: Name of field to add
    - `value`: Numeric (float) value of new field
    '''

    bl = get_beeline()

    if bl:
        bl.tracer_impl.add_rollup_field(name=name, value=value)


def add_trace_field(name, value):
    ''' Similar to `add_context_field` - adds a field to the current span, but
    also to all other future spans in this trace. Trace context fields will be
    propagated to downstream services if using instrumented libraries
    like `requests`.

    Args:
    - `name`: Name of field to add
    - `value`: Value of new field
    '''
    bl = get_beeline()

    if bl:
        bl.tracer_impl.add_trace_field(name=name, value=value)


def remove_trace_field(name):
    ''' Removes a trace context field from the current span. This will not
    affect  other existing spans, but will prevent the field from being
    propagated to new spans.

    Args:
    - `name`: Name of field to remove
    '''

    bl = get_beeline()

    if bl:
        bl.tracer_impl.remove_trace_field(name=name)


def tracer(name, trace_id=None, parent_id=None):
    '''
    When used in a context manager, creates a span for the contained
    code. If a trace is ongoing, will add a new child span under the currently
    running span. If no trace is ongoing, will start a new trace.

    Example use:

    ```
    with tracer(name="my expensive computation"):
        recursive_fib(100)
    ```

    Args:
    - `name`: a descriptive name for the this trace span, i.e. "database query for user"
    - `trace_id`: the trace_id to use. If None, will be automatically generated if no
       current trace is ongoing. Use this if you want to explicitly resume a trace
       in this application that was initiated in another application, and you have
       the upstream trace_id.
    - `parent_id`: If trace_id is set, will populate the root span's parent
        with this id.
    '''
    bl = get_beeline()
    if bl:
        return bl.tracer(name=name, trace_id=trace_id, parent_id=parent_id)

    # if the beeline is not initialized, build a dummy function
    # that will work as a context manager and call that
    @contextmanager
    def _noop_cm():
        yield

    return _noop_cm()


def start_trace(context=None, trace_id=None, parent_span_id=None):
    '''
    Start a trace, returning the root span. To finish the trace, pass the span
    to `finish_trace`. `start_trace` does not propagate contexts - if you wish
    to propagate contexts from sources such as HTTP headers, use `propagate_and_start_trace`
    instead. If you are using the beeline middleware plugins, such as fordjango,
    flask, or AWS lambda, you will want to use `start_span` instead, as `start_trace`
    is called at the start of the request.

    Args:
    - `context`: optional dictionary of event fields to populate the root span with
    - `trace_id`: the trace_id to use. If None, will be automatically generated.
        Use this if you want to explicitly resume trace in this application that was
        initiated in another application, and you have the upstream trace_id.
    - `parent_span_id`: If trace_id is set, will populate the root span's parent
        with this id.
    '''
    bl = get_beeline()

    if bl:
        return bl.tracer_impl.start_trace(context=context, trace_id=trace_id, parent_span_id=parent_span_id)


def finish_trace(span):
    ''' Explicitly finish a trace. If you started a trace with `start_trace`, you must call
    this to close the trace and send the root span. If you are using the beeline middleware plugins,
    such as django, flask, or AWS lambda, you can skip this step as the trace will be closed for
    you.

    Args:
    - `span`: Span object that was returned by `start_trace`
    '''
    bl = get_beeline()

    if bl:
        bl.tracer_impl.finish_trace(span=span)


def start_span(context=None, parent_id=None):
    '''
    Start a new span and return the span object. Returns None if no trace is active.
    For each `start_span`, there should be one call to `close_span`. Child spans should
    also be closed before parent spans. Closing spans out of order will lead to strange
    results and can break the bookkeeping needed to preserve trace structure. For example:

    ```
    parent_span = beeline.start_span()
    # this span is a child of the last span created
    child_span = beeline.start_span()
    beeline.finish_span(child_span)
    beeline.finish_span(parent_span)
    ```

    Args:
    - `context`: optional dictionary of event fields to populate the span with
    - `parent_id`: ID of parent span - use this only if you have a very good reason to
        do so.
    '''
    bl = get_beeline()

    if bl:
        return bl.tracer_impl.start_span(context=context, parent_id=parent_id)


def finish_span(span):
    '''
    Finish the provided span, sending the associated event data to Honeycomb.

    For each `start_span`, there should be one call to `finish_span`.

    Args:
    - `span`: Span object that was returned by `start_trace`
    '''

    bl = get_beeline()

    if bl:
        bl.tracer_impl.finish_span(span=span)


def propagate_and_start_trace(context, request):
    '''
    Given context and a beeline.propagation.Request subclass, calls the header_parse hooks
    to propagate information from the incoming http (or similar) request context,
    returning a new trace using that information if it exists.
    '''
    bl = get_beeline()

    if bl:
        return bl.tracer_impl.propagate_and_start_trace(context, request)
    return None


def http_trace_parser_hook(headers):
    '''
    Given headers, calls the header_parse hooks to propagate information from the
    incoming http (or similar) request context, returning a new trace using that
    information if it exists.
    '''
    bl = get_beeline()

    if bl:
        return bl.tracer_impl.http_trace_parser_hook(headers)
    return None


def http_trace_propagation_hook():
    '''
    Given headers, calls the header_parse hooks to propagate information from the
    incoming http (or similar) request context, returning a new trace using that
    information if it exists.
    '''
    bl = get_beeline()

    if bl:
        try:
            return bl.tracer_impl.http_trace_propagation_hook(bl.tracer_impl.get_propagation_context())
        except Exception:
            err = sys.exc_info()
            bl.log('error: http_trace_propagation_hook returned exception: %s', err)
    return None


def marshal_trace_context():
    '''
    DEPRECATED: Returns a serialized form of the current trace context (including the trace
    id and the current span), encoded as a string. You can use this to propagate
    trace context to other services.

    Use `beeline.propagation.honeycomb` functions to work with honeycomb trace context instead.

    Example:

    ```
    trace_context = beeline.marshal_trace_context()
    headers = {'X-Honeycomb-Trace': trace_context}
    requests.get("http://...", headers=headers)
    ```
    '''
    bl = get_beeline()

    if bl:
        return bl.tracer_impl.marshal_trace_context()
    return None


def new_event(data=None, trace_name=''):
    ''' DEPRECATED: Helper method that wraps `start_trace` and
    `start_span`. It is better to use these methods as it provides
    better control and context around how traces are implemented in your
    app.

    Creates a new span, populating it with the given data if
    supplied. If no trace is running, a new trace will be started,
    otherwise the event will be added as a span of the existing trace.

    To send the event, call `beeline.send_event()`. There should be a
    `send_event()` for each call to `new_event()`, or tracing and
    `add` and `add_field` will not work correctly.

    If trace_name is specified, will set the "name" field of the current span,
    which is used in the trace visualizer.
    '''

    bl = get_beeline()

    if bl:
        bl.new_event(data=data, trace_name=trace_name)


def send_event():
    ''' DEPRECATED: Sends the currently active event (current span),
    if it exists.

    There must be one call to `send_event` for each call to `new_event`.
    '''

    bl = get_beeline()

    if bl:
        bl.send_event()


def send_all():
    ''' send all spans in the trace stack, regardless of their
    state. You might use this in a catch-all error handler
    along with `beeline.close()` to send all events before the program
    terminates abruptly.
    '''

    bl = get_beeline()

    if bl:
        bl.send_all()


def get_beeline():
    return _GBL


def get_responses_queue():
    '''
    Returns a queue from which you can read a record of response info from
    each event sent. Responses will be dicts with the following keys:
        - `status_code` - the HTTP response from the api (eg. 200 or 503)
        - `duration` - how long it took to POST this event to the api, in ms
        - `metadata` - pass through the metadata you added on the initial event
        - `body` - the content returned by API (will be empty on success)
        - `error` - in an error condition, this is filled with the error message
    When the Client's `close` method is called, a None will be inserted on
    the queue, indicating that no further responses will be written.
    '''

    bl = get_beeline()

    if bl:
        return bl.get_responses_queue()


def close():
    ''' close the beeline and libhoney client, flushing any unsent events. '''
    global _GBL
    if _GBL:
        _GBL.close()

    _GBL = None


def traced(name, trace_id=None, parent_id=None):
    '''
    Function decorator to wrap an entire function in a trace span. If no trace
    is active in the current thread, starts a new trace, and the wrapping span
    will be a root span. If a trace is active, creates a child span of the
    existing trace.

    Example use:

    ```
    @traced(name="my_expensive_function")
    def my_func(n):
        recursive_fib(n)

    my_func(100)
    ```

    Args:
    - `name`: a descriptive name for the this trace span, i.e. "function_name". This is required.
    - `trace_id`: the trace_id to use. If None, will be automatically generated.
        Use this if you want to explicitly resume a trace in this application that was
        initiated in another application, and you have the upstream trace_id.
    - `parent_id`: If trace_id is set, will populate the root span's parent
        with this id.
    '''

    return traced_impl(tracer_fn=tracer, name=name, trace_id=trace_id, parent_id=parent_id)


def traced_thread(fn):
    '''
    Function decorator to pass context to a function that is a thread target. Because the beeline uses
    thread-local storage to keep track of trace state, tracing doesn't work across threads unless the state
    is explicitly passed between threads. You can use this decorator to more easily pass state to a thread.

    Example use:

    ```
    @traced(name="my_async_function")
    def my_func():
        # ...
        with beeline.tracer(name="do_stuff"):
            do_stuff()

    # we want to call my_func asynchronously by passing to a thread or a thread pool
    @beeline.traced_thread
    def _my_func_t():
        return my_func()

    t = threading.Thread(target=_my_func_t)
    t.start()
    ```

    '''

    # if beeline is not initialized, or there is no active trace, do nothing
    bl = get_beeline()
    if bl is None or bl.tracer_impl.get_active_trace_id() is None:
        @functools.wraps(fn)
        def noop(*args, **kwargs):
            return fn(*args, **kwargs)
        return noop

    trace_copy = bl.tracer_impl._trace.copy()

    @functools.wraps(fn)
    def wrapped(*args, **kwargs):
        bl.tracer_impl._trace = trace_copy
        return fn(*args, **kwargs)

    return wrapped



import threading
import unittest
from mock import Mock, patch, call

import beeline
import libhoney
assert libhoney


class TestBeeline(unittest.TestCase):
    def setUp(self):
        self.addCleanup(patch.stopall)
        self.m_gbl = patch('beeline._GBL').start()

    def test_send_event(self):
        ''' test correct behavior for send_event '''
        _beeline = beeline.Beeline()
        _beeline.tracer_impl = Mock()
        m_span = Mock()
        _beeline.tracer_impl.get_active_span.return_value = m_span
        _beeline.send_event()
        _beeline.tracer_impl.get_active_span.assert_called_once_with()
        _beeline.tracer_impl.finish_trace.assert_called_once_with(m_span)

    def test_send_no_events(self):
        ''' ensure nothing crashes when we try to send with no events in the
        stack '''
        _beeline = beeline.Beeline()
        _beeline.tracer_impl = Mock()
        _beeline.tracer_impl.get_active_span.return_value = None
        _beeline.send_event()
        _beeline.tracer_impl.get_active_span.assert_called_once_with()

    def test_send_all(self):
        ''' ensure events are flushed, and that the root span is handled with
        finish_trace '''
        s1, s2, s3 = Mock(), Mock(), Mock()
        s1.is_root.return_value = False
        s2.is_root.return_value = False
        s3.is_root.return_value = True
        _beeline = beeline.Beeline()
        _beeline.tracer_impl = Mock()
        _beeline.tracer_impl.get_active_span.side_effect = [s1, s2, s3, None]

        _beeline.send_all()

        _beeline.tracer_impl.finish_span.assert_has_calls([
            call(s1),
            call(s2),
        ])
        _beeline.tracer_impl.finish_trace.assert_called_once_with(s3)

    def test_run_hooks_and_send_no_hooks(self):
        ''' ensure send works when no hooks defined '''
        ev = Mock()
        _beeline = beeline.Beeline()
        _beeline.tracer_impl = Mock()
        _beeline._run_hooks_and_send(ev)

        # no hooks, not a traced event - call send
        ev.send.assert_called_once_with()
        ev.send_presampled.assert_not_called()

    def test_run_hooks_and_send_sampler(self):
        ''' ensure send works with a sampler hook defined '''
        def _sampler_drop_all(fields):
            return False, 0
        m_sampler_hook = Mock()
        m_sampler_hook.side_effect = _sampler_drop_all
        _beeline = beeline.Beeline(sampler_hook=m_sampler_hook)
        _beeline.tracer_impl = Mock()
        ev = Mock()

        _beeline._run_hooks_and_send(ev)
        m_sampler_hook.assert_called_once_with(ev.fields())
        ev.send_presampled.assert_not_called()
        ev.send.assert_not_called()

        def _sampler_drop_none(fields):
            return True, 100

        ev = Mock()
        m_sampler_hook.reset_mock()

        m_sampler_hook.side_effect = _sampler_drop_none

        _beeline._run_hooks_and_send(ev)
        m_sampler_hook.assert_called_once_with(ev.fields())
        self.assertEqual(ev.sample_rate, 100)
        ev.send_presampled.assert_called_once_with()
        ev.send.assert_not_called()

    def test_run_hooks_and_send_presend_hook(self):
        ''' ensure send works when presend hook is defined '''
        def _presend_hook(fields):
            fields["thing i want"] = "put it there"
            del fields["thing i don't want"]
        m_presend_hook = Mock()
        m_presend_hook.side_effect = _presend_hook
        _beeline = beeline.Beeline(presend_hook=m_presend_hook)
        _beeline.tracer_impl = Mock()

        ev = Mock()
        ev.fields.return_value = {
            "thing i don't want": "get it out of here",
            "happy data": "so happy",
        }

        _beeline._run_hooks_and_send(ev)
        ev.send_presampled.assert_not_called()
        ev.send.assert_called_once_with()
        self.assertDictEqual(
            ev.fields(),
            {
                "thing i want": "put it there",
                "happy data": "so happy",
            },
        )

    def test_start_trace_returns_value(self):
        ''' ensure the top-level start_span and start_trace APIs return the value
        form their calls to the tracer '''
        self.m_gbl.tracer_impl.start_span.return_value = 'wooimaspan'
        val = beeline.start_span()
        self.assertEqual(val, 'wooimaspan')

        self.m_gbl.tracer_impl.start_trace.return_value = 'wooimatrace'
        val = beeline.start_trace()
        self.assertEqual(val, 'wooimatrace')

    def test_marshal_trace_context_returns_value(self):
        ''' ensure the top-level definition of marshal_trace_context returns a value '''
        self.m_gbl.tracer_impl.marshal_trace_context.return_value = 'asdf'
        val = beeline.marshal_trace_context()
        self.assertEqual(val, 'asdf')

    def test_trace_wrapper(self):
        ''' ensure that the trace wrapper decorates a function and starts a trace '''
        _beeline = beeline.Beeline()
        with patch('beeline.get_beeline') as m_gbl:
            m_gbl.return_value = _beeline
            _beeline.tracer_impl._run_hooks_and_send = Mock()

            @beeline.traced(name="my_sum")
            def my_sum(a, b):
                return a + b

            # this should accept the function's arguments normally and return the function's value
            # if there is one
            self.assertEqual(my_sum(1, 2), 3)
            # check that an event was sent, from which we can infer that the function was wrapped
            self.assertTrue(_beeline.tracer_impl._run_hooks_and_send.called)

    def test_generator_wrapper(self):
        ''' ensure that the trace wrapper decorates a generator function and starts a trace
            also ensure that child traces get the parent trace correctly set
        '''

        _beeline = beeline.Beeline()

        with patch('beeline.get_beeline') as m_gbl:
            m_gbl.return_value = _beeline
            _beeline.tracer_impl._run_hooks_and_send = Mock()

            @beeline.traced(name="return_integer_n")
            def return_integer(n):
                return n

            @beeline.traced(name="output_integers_to")
            def output_integers_to(n):
                for i in range(n):
                    yield return_integer(i)

            # should accept the function's arguments normally and yield the items from the
            # generator
            self.assertEqual(list(output_integers_to(3)), [0, 1, 2])

            self.assertTrue(_beeline.tracer_impl._run_hooks_and_send.called)

            spans = [x[0][0]
                     for x in _beeline.tracer_impl._run_hooks_and_send.call_args_list]

            # check the child spans now
            parent_span = spans[-1]
            child_spans = spans[:-1]

            for child_span in child_spans:
                self.assertEqual(child_span.parent_id, parent_span.id)

    @staticmethod
    def raising_run_in_thread(target):
        closure_dict = {}

        def wrapper():
            try:
                target()
            except Exception as exc:
                closure_dict["thread_exception"] = exc

        thread = threading.Thread(target=wrapper)
        thread.start()
        thread.join()

        if "thread_exception" in closure_dict:
            raise closure_dict["thread_exception"]

    def test_threaded_trace(self):
        _beeline = beeline.Beeline()

        with patch('beeline.get_beeline') as m_gbl:
            m_gbl.return_value = _beeline
            _beeline.tracer_impl._run_hooks_and_send = Mock()

            _beeline.tracer_impl.start_trace(trace_id="asdf")
            self.assertEqual(_beeline.tracer_impl._trace.id, "asdf")

            def thread_func():
                # confirm no trace state in new thread
                self.assertIsNone(_beeline.tracer_impl._trace)

            self.raising_run_in_thread(target=thread_func)

            @beeline.traced_thread
            def traced_thread_func():
                self.assertEqual(_beeline.tracer_impl._trace.id, "asdf")

                with _beeline.tracer(name="foo") as span:
                    self.assertEqual(span.trace_id, "asdf")
                    self.assertEqual(
                        span.parent_id, _beeline.tracer_impl._trace.stack[0].id)

            self.raising_run_in_thread(target=traced_thread_func)

            # test use of beeline client
            @_beeline.traced_thread
            def traced_thread_func_2():
                self.assertEqual(_beeline.tracer_impl._trace.id, "asdf")

                with _beeline.tracer(name="foo2") as span:
                    self.assertEqual(span.trace_id, "asdf")
                    self.assertEqual(
                        span.parent_id, _beeline.tracer_impl._trace.stack[0].id)

            self.raising_run_in_thread(target=traced_thread_func_2)

    def test_finish_span_none(self):
        ''' ensure finish_span does not crash if an empty span is passed to it '''
        _beeline = beeline.Beeline()
        # should not crash
        _beeline.tracer_impl.finish_span(None)


class TestBeelineNotInitialized(unittest.TestCase):
    def setUp(self):
        self.addCleanup(patch.stopall)
        self.m_gbl = patch('beeline.get_beeline').start()
        self.m_gbl.return_value = None

    def test_trace_wrapper(self):
        ''' ensure the trace wrapper doesn't break if the beeline is not initialized '''
        self.assertIsNone(beeline.get_beeline())

        @beeline.traced(name="my_sum")
        def my_sum(a, b):
            return a + b

        # this should not crash if the beeline isn't initialized
        # it should also accept arguments normally and return the function's value
        self.assertEqual(my_sum(1, 2), 3)

    def test_tracer_context_manager(self):
        ''' ensure the tracer context manager doesn't break if the beeline is not initialized '''
        self.assertIsNone(beeline.get_beeline())

        def my_sum(a, b):
            with beeline.tracer(name="my_sum"):
                return a + b

        # this should not crash if the beeline isn't initialized
        # it should also accept arguments normally and return the function's value
        self.assertEqual(my_sum(1, 2), 3)

    def test_traced_thread(self):
        self.assertIsNone(beeline.get_beeline())

        @beeline.traced_thread
        def my_sum(a, b):
            return a + b

        # this should not crash if the beeline isn't initialized
        # it should also accept arguments normally and return the function's value
        self.assertEqual(my_sum(1, 2), 3)



from __future__ import absolute_import
import unittest
from mock import Mock, patch

import jinja2
import beeline.patch.jinja2
assert beeline.patch.jinja2  # make pyflake stop complainings


class TestJinja2Patch(unittest.TestCase):
    def test_wrapper_executes(self):

        with patch('beeline.patch.jinja2.beeline') as m_beeline:
            m_span = Mock()
            m_beeline.start_span.return_value = m_span

            t = jinja2.Template("my template")
            t.render()
            m_beeline.start_span.assert_called_once_with(context={
                "name": "jinja2_render_template",
                "template.name": "[string]",
            })
            m_beeline.finish_span.assert_called_once_with(m_span)

            m_beeline.reset_mock()
            m_beeline.start_span.return_value = m_span

            t = jinja2.Template("my template")
            t.name = 'foo'
            t.render()
            m_beeline.start_span.assert_called_once_with(context={
                "name": "jinja2_render_template",
                "template.name": "foo",
            })
            m_beeline.finish_span.assert_called_once_with(m_span)



from wrapt import wrap_function_wrapper
import beeline


def _render_template(fn, instance, args, kwargs):
    span = beeline.start_span(context={
        "name": "jinja2_render_template",
        "template.name": instance.name or "[string]",
    })

    try:
        return fn(*args, **kwargs)
    finally:
        beeline.finish_span(span)


wrap_function_wrapper('jinja2', 'Template.render', _render_template)



import datetime
import threading

import beeline
import flask  # to avoid namespace collision with request vs Request
from beeline.propagation import Request
from flask import current_app, signals
# needed to build a request object from environ in the middleware
from werkzeug.wrappers import Request
from beeline.middleware.wsgi import WSGIRequest


class HoneyMiddleware(object):

    def __init__(self, app, db_events=True):
        self.app = app
        self.app.before_request(self._before_request)
        if signals.signals_available:
            self.app.teardown_request(self._teardown_request)
        app.wsgi_app = HoneyWSGIMiddleware(app.wsgi_app)
        if db_events:
            app = HoneyDBMiddleware(app)

    def _before_request(self):
        beeline.add_field("request.route", flask.request.endpoint)

    def _teardown_request(self, exception):
        if exception:
            beeline.add_field('request.error_detail',
                              beeline.internal.stringify_exception(exception))
            beeline.add_field('request.error', str(type(exception)))

        beeline.internal.send_event()


class HoneyWSGIMiddleware(object):

    def __init__(self, app):
        self.app = app

    def __call__(self, environ, start_response):
        req = Request(environ, shallow=True)
        wr = WSGIRequest("flask", environ)

        root_span = beeline.propagate_and_start_trace(wr.request_context(), wr)

        def _start_response(status, headers, *args):
            status_code = int(status[0:4])
            beeline.add_context_field("response.status_code", status_code)
            if not signals.signals_available:
                beeline.finish_trace(root_span)

            return start_response(status, headers, *args)

        return self.app(environ, _start_response)


class HoneyDBMiddleware(object):

    def __init__(self, app=None):
        self.app = app
        if app is not None:
            self.init_app(app)

        self.state = threading.local()
        self.state.span = None

    def init_app(self, app):
        try:
            from sqlalchemy.engine import Engine  # pylint: disable=bad-option-value,import-outside-toplevel
            from sqlalchemy.event import listen  # pylint: disable=bad-option-value,import-outside-toplevel

            listen(Engine, 'before_cursor_execute', self.before_cursor_execute)
            listen(Engine, 'after_cursor_execute', self.after_cursor_execute)
            listen(Engine, 'handle_error', self.handle_error)
        except ImportError:
            pass

    def before_cursor_execute(self, conn, cursor, statement, parameters, context, executemany):
        if not current_app:
            return

        params = []

        # the type of parameters passed in varies depending on DB - handle list, dict, and tuple
        if type(parameters) == tuple or type(parameters) == list:
            for param in parameters:
                if type(param) == datetime.datetime:
                    param = param.isoformat()
                params.append(param)
        elif type(parameters) == dict:
            for k, v in parameters.items():
                param = "%s=" % k
                if type(v) == datetime.datetime:
                    v = v.isoformat()
                param += str(v)
                params.append(param)

        self.state.span = beeline.start_span(context={
            "name": "flask_db_query",
            "type": "db",
            "db.query": statement,
            "db.query_args": params,
        })

        self.query_start_time = datetime.datetime.now()

    def after_cursor_execute(self, conn, cursor, statement, parameters, context, executemany):
        if not current_app:
            return

        query_duration = datetime.datetime.now() - self.query_start_time

        beeline.add_context({
            "db.duration": query_duration.total_seconds() * 1000,
            "db.last_insert_id": getattr(cursor, 'lastrowid', None),
            "db.rows_affected": cursor.rowcount,
        })
        if self.state.span:
            beeline.finish_span(self.state.span)
        self.state.span = None

    def handle_error(self, context):
        if not current_app:
            return

        beeline.add_context_field(
            "db.error", beeline.internal.stringify_exception(context.original_exception))
        if self.state.span:
            beeline.finish_span(self.state.span)
        self.state.span = None



import beeline
from wrapt import wrap_function_wrapper
import requests
# needed for pyflakes
assert requests


def request(_request, instance, args, kwargs):
    span = beeline.start_span(context={"meta.type": "http_client"})

    b = beeline.get_beeline()
    if b and b.http_trace_propagation_hook is not None:
        new_headers = beeline.http_trace_propagation_hook()
        if new_headers:
            b.log(
                "requests lib - adding trace context to outbound request: %s", new_headers)
            instance.headers.update(new_headers)
        else:
            b.log("requests lib - no trace context found")

    try:
        resp = None

        # Required as Python treats the `or` keyword differently in string
        # interpolation vs. when assigning a variable.
        method = kwargs.get('method') or args[0]

        beeline.add_context({
            "name": "requests_%s" % method,
            "request.method": method,
            "request.url": kwargs.get('url') or args[1],
        })
        resp = _request(*args, **kwargs)
        return resp
    except Exception as e:
        beeline.add_context({
            "request.error_type": str(type(e)),
            "request.error": beeline.internal.stringify_exception(e),
        })
        raise
    finally:
        if resp is not None:
            content_type = resp.headers.get('content-type')
            if content_type:
                beeline.add_context_field(
                    "response.content_type", content_type)
            content_length = resp.headers.get('content-length')
            if content_length:
                beeline.add_context_field(
                    "response.content_length", content_length)
            if hasattr(resp, 'status_code'):
                beeline.add_context_field(
                    "response.status_code", resp.status_code)
        beeline.finish_span(span)


wrap_function_wrapper('requests.sessions', 'Session.request', request)



import unittest
from beeline.propagation import DictRequest, PropagationContext
import beeline.propagation.honeycomb as hc
import beeline.propagation.w3c as w3c

_TEST_TRACEPARENT_HEADER = "00-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-00"
_TEST_TRACEPARENT_HEADER_DEFAULT_TRACEFLAGS = "00-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-01"
_TEST_TRACESTATE_HEADER = "foo=bar,bar=baz"

_TEST_HEADERS = {
    "traceparent": _TEST_TRACEPARENT_HEADER,
    "tracestate": _TEST_TRACESTATE_HEADER
}
_TEST_TRACE_ID = "0af7651916cd43dd8448eb211c80319c"
_TEST_PARENT_ID = "b7ad6b7169203331"
_TEST_TRACE_FLAGS = "00"
_TEST_TRACESTATE = "foo=bar,bar=baz"


class TestW3CMarshalUnmarshal(unittest.TestCase):
    def test_roundtrip(self):
        '''Verify that we can successfully roundtrip (marshal and unmarshal)'''
        trace_fields = {"traceflags": _TEST_TRACE_FLAGS,
                        "tracestate": _TEST_TRACESTATE}

        pc = PropagationContext(_TEST_TRACE_ID, _TEST_PARENT_ID, trace_fields)

        traceparent_header = w3c.marshal_traceparent(pc)
        tracestate_header = w3c.marshal_tracestate(pc)

        # Make sure marshalled headers are as we expect.
        self.assertEqual(_TEST_TRACEPARENT_HEADER, traceparent_header)
        self.assertEqual(_TEST_TRACESTATE_HEADER, tracestate_header)

        new_trace_id, new_parent_id, new_trace_flags = w3c.unmarshal_traceparent(
            traceparent_header)
        new_tracestate = w3c.unmarshal_tracestate(tracestate_header)

        # Check round-trip values are the same as start values.
        self.assertEqual(_TEST_TRACE_ID, new_trace_id)
        self.assertEqual(_TEST_PARENT_ID, new_parent_id)
        self.assertEqual(_TEST_TRACE_FLAGS, new_trace_flags)
        self.assertEqual(_TEST_TRACESTATE, new_tracestate)


class TestW3CHTTPTraceParserHook(unittest.TestCase):
    def test_has_header(self):
        '''Test that the hook properly parses W3C trace headers'''
        req = DictRequest(_TEST_HEADERS)
        pc = w3c.http_trace_parser_hook(req)
        self.assertEqual(pc.trace_id, _TEST_TRACE_ID)
        self.assertEqual(pc.parent_id, _TEST_PARENT_ID)
        self.assertEqual(pc.trace_fields, {
            "tracestate": _TEST_TRACESTATE,
            "traceflags": _TEST_TRACE_FLAGS
        })

    def test_no_header(self):
        req = DictRequest({})
        pc = w3c.http_trace_parser_hook(req)
        self.assertIsNone(pc)


class TestW3CHTTPTracePropagationHook(unittest.TestCase):
    def test_generates_correct_headers(self):
        pc = PropagationContext(
            _TEST_TRACE_ID, _TEST_PARENT_ID, {"traceflags": _TEST_TRACE_FLAGS,
                                              "tracestate": _TEST_TRACESTATE}
        )
        headers = w3c.http_trace_propagation_hook(pc)
        self.assertIn('traceparent', headers)
        self.assertIn('tracestate', headers)
        self.assertEqual(headers['traceparent'],
                         _TEST_TRACEPARENT_HEADER)
        self.assertEqual(headers['tracestate'],
                         _TEST_TRACESTATE_HEADER)

    def test_defaults_trace_flags_to_sampled(self):
        pc = PropagationContext(
            _TEST_TRACE_ID, _TEST_PARENT_ID
        )
        headers = w3c.http_trace_propagation_hook(pc)
        self.assertIn('traceparent', headers)
        self.assertEqual(headers['traceparent'],
                         _TEST_TRACEPARENT_HEADER_DEFAULT_TRACEFLAGS)



from beeline.propagation import Request


class WSGIRequest(Request):
    def __init__(self, mwname, environ):
        self._mwname = mwname
        self._environ = environ

    def header(self, key):
        # FIXME: Is this .upper strictly necessary? Does environ already do it for us?
        lookup_key = "HTTP_" + key.upper().replace('-', '_')
        return self._environ.get(lookup_key)

    def method(self):
        return self._environ.get('REQUEST_METHOD')

    def scheme(self):
        return self._environ.get('wsgi.url_scheme')

    def host(self):
        return self._environ.get('HTTP_HOST')

    def path(self):
        return self._environ.get('PATH_INFO')

    def query(self):
        return self._environ.get('QUERY_STRING')

    def middleware_request(self):
        return self._environ

    def request_context(self):
        request_method = self._environ.get('REQUEST_METHOD')
        if request_method:
            trace_name = "%s_http_%s" % (self._mwname, request_method.lower())
        else:
            trace_name = "%s_http" % self._mwname

        return {
            "name": trace_name,
            "type": "http_server",
            "request.host": self._environ.get('HTTP_HOST'),
            "request.method": request_method,
            "request.path": self._environ.get('PATH_INFO'),
            "request.remote_addr": self._environ.get('REMOTE_ADDR'),
            "request.content_length": self._environ.get('CONTENT_LENGTH', 0),
            "request.user_agent": self._environ.get('HTTP_USER_AGENT'),
            "request.scheme": self._environ.get('wsgi.url_scheme'),
            "request.query": self._environ.get('QUERY_STRING')
        }



import six
from abc import ABCMeta, abstractmethod
import beeline


def propagate_dataset():
    return bool


class PropagationContext(object):
    '''
    PropagationContext represents information that can either be read from, or
    propagated via standard trace propagation headers such as X-Honeycomb-Trace
    and W3C headers. http_trace_parser_hooks generate this from requests, and
    http_trace_propagation_hooks use this to generate a set of headers for
    outbound HTTP requests.
    '''

    def __init__(self, trace_id, parent_id, trace_fields={}, dataset=None):
        self.trace_id = trace_id
        self.parent_id = parent_id
        self.trace_fields = trace_fields
        self.dataset = dataset


@six.add_metaclass(ABCMeta)
class Request(object):
    '''
    beeline.propagation.Request is an abstract class that defines the interface that should
    be used by middleware to pass request information into http_trace_parser_hooks. It should
    at a minimum contain the equivalent of HTTP headers, and optionally include other HTTP
    information such as method, schema, host, path, and query. The `middleware_request` method
    returns a middleware-specific request object or equivalent that can be used by custom hooks
    to extract additional information to be added to the trace or propagated.
    '''

    @abstractmethod
    def header(self, key):
        """
        Get the value associated with the specified key, transformed as necessary for the
        transport and middleware.
        """
        pass

    @abstractmethod
    def method(self):
        """
        For HTTP requests, the HTTP method (GET, PUT, etc.) of the request.
        """
        pass

    @abstractmethod
    def scheme(self):
        """
        For HTTP requests, the scheme (http, https, etc.)
        """
        pass

    @abstractmethod
    def host(self):
        """
        For HTTP requests, the host part of the URL (e.g. www.honeycomb.io, api.honeycomb.io:80)
        """
        pass

    @abstractmethod
    def path(self):
        """
        For HTTP requests, the path part of the URL (e.g. /1/event/)
        """
        pass

    @abstractmethod
    def query(self):
        """
        For HTTP requests, the query part of the URL (e.g. key1=value1&key2=value2)
        """
        pass

    @abstractmethod
    def middleware_request(self):
        """
        The middleware-specific source of request data (for middleware-specific custom hooks)
        """
        pass


class DictRequest(Request):
    '''
    Basic dictionary request that just takes in a dictionary of headers, and a dictionary
    of request properties. Primarily for testing.
    '''

    def __init__(self, headers, props={}):
        self._headers = headers
        self._props = props
        self._keymap = {k.lower(): k for k in self._headers.keys()}

    def header(self, key):
        lookup_key = key.lower()
        if lookup_key not in self._keymap:
            return None
        lookup_key = self._keymap[lookup_key]
        return self._headers[lookup_key]

    def method(self):
        return self._props['method']

    def scheme(self):
        return self._props['scheme']

    def host(self):
        return self._props['host']

    def path(self):
        return self._props['path']

    def query(self):
        return self._props['query']

    def middleware_request(self):
        return {
            'headers': self._headers,
            'props': self._props
        }



import unittest
from mock import Mock, patch

import beeline


class TestRequestsPatch(unittest.TestCase):
    def test_request_fn_injects_headers_and_returns(self):
        from beeline.patch.requests import request  # pylint: disable=bad-option-value,import-outside-toplevel

        with patch('beeline.get_beeline') as m_bl:
            bl = Mock()
            m_bl.return_value = bl

            trace_context = "1;trace_id=foo,parent_id=bar,context=base64value=="

            bl.tracer_impl.http_trace_propagation_hook.return_value = {
                'X-Honeycomb-Trace': trace_context
            }

            # this is the class instance (Session object) that is passed to our request function
            # by wrapt
            m_session = Mock()
            m_session.headers = {}

            # this is our request call that's being wrapped
            m_request = Mock()
            m_request.return_value = Mock(
                headers={'content-type': 'application/json', 'content-length': 23}, status_code=500)
            args = ['get']
            kwargs = {'url': 'http://example.com'}
            ret = request(m_request, m_session, args, kwargs)

            m_request.assert_called_once_with(*args, **kwargs)
            self.assertEqual(ret, m_request.return_value)
            self.assertEqual(
                m_session.headers['X-Honeycomb-Trace'], trace_context)

    def test_request_fn_injects_correct_context(self):
        ''' confirm that the injected trace context references the child span wrapping the request call '''
        from beeline.patch.requests import request  # pylint: disable=bad-option-value,import-outside-toplevel

        _beeline = beeline.Beeline(transmission_impl=Mock())
        # prevent spans from being sent/closed so we can get at the span generated by the request patch
        _beeline.tracer_impl.finish_span = Mock()

        with patch('beeline.get_beeline') as m_bl:
            m_bl.return_value = _beeline

            with _beeline.tracer("parent", trace_id="abc"):
                parent_span = _beeline.tracer_impl.get_active_span()

                # this is the class instance (Session object) that is passed to our request function by wrapt
                m_session = Mock()
                m_session.headers = {}

                # this is our request call that's being wrapped
                m_request = Mock()
                m_request.return_value = Mock(
                    headers={'content-type': 'application/json', 'content-length': 23}, status_code=500)
                args = ['get']
                kwargs = {'url': 'http://example.com'}
                ret = request(m_request, m_session, args, kwargs)

                m_request.assert_called_once_with(*args, **kwargs)
                self.assertEqual(ret, m_request.return_value)

                self.assertEqual(len(_beeline.tracer_impl._trace.stack), 2)
                child_span = _beeline.tracer_impl.get_active_span()
                # we should have two distinct spans in this trace
                self.assertNotEqual(child_span.id, parent_span.id)
                self.assertEqual(child_span.parent_id, parent_span.id)

                trace_context = m_session.headers['X-Honeycomb-Trace']
                trace_id, parent_id, _ = beeline.trace.unmarshal_trace_context(
                    trace_context)
                # confirm the trace context parent is the child span, not the parent span
                self.assertEqual(child_span.trace_id, trace_id)
                self.assertEqual(child_span.id, parent_id)
                # should be the same trace
                self.assertEqual(parent_span.trace_id, trace_id)

    def test_request_fn_works_without_init(self):
        ''' verify that the requests wrapper works even if the beeline is not initialized '''
        from beeline.patch.requests import request  # pylint: disable=bad-option-value,import-outside-toplevel

        # this is the class instance (Session object) that is passed to our request function by wrapt
        m_session = Mock()
        m_session.headers = {}

        m_request = Mock()
        m_request.return_value = Mock(
            headers={'content-type': 'application/json', 'content-length': 23}, status_code=500)
        args = ['get']
        kwargs = {'url': 'http://example.com'}
        ret = request(m_request, m_session, args, kwargs)

        m_request.assert_called_once_with(*args, **kwargs)
        self.assertEqual(ret, m_request.return_value)



from collections import defaultdict
from mock import ANY, Mock, call, patch
import base64
import datetime
import json
import unittest
import re
import os
import binascii
import beeline.propagation

from libhoney import Event

from beeline.trace import (
    _should_sample, SynchronousTracer, marshal_trace_context,
    unmarshal_trace_context, Span, generate_span_id, generate_trace_id
)

from beeline.propagation import DictRequest


class TestIDGeneration(unittest.TestCase):
    def test_span_id(self):
        '''Test that span ID generation meets basic length and character constraints'''
        span_id = generate_span_id()
        self.assertIsNotNone(re.match(r"[\da-f]{16}", span_id))

    def test_trace_id(self):
        '''Test that trace ID generation meets basic length and character constraints'''
        trace_id = generate_trace_id()
        self.assertIsNotNone(re.match(r"[\da-f]{32}", trace_id))


class TestTraceSampling(unittest.TestCase):
    def test_deterministic(self):
        ''' test a specific id that should always work with the given sample rate '''
        trace_id = 'b8d674f1-04ed-4ea8-b16d-b4dbbe87c78e'
        n = 0
        while n < 1000:
            self.assertTrue(_should_sample(trace_id, 1000))
            n += 1

    def test_deterministic_interop(self):
        '''test a specific set of ids that should always have the given result (at sample rate 2)
        Ensures interoperability with other beelines'''
        ids = [
            "4YeYygWjTZ41zOBKUoYUaSVxPGm78rdU",
            "iow4KAFBl9u6lF4EYIcsFz60rXGvu7ph",
            "EgQMHtruEfqaqQqRs5nwaDXsegFGmB5n",
            "UnVVepVdyGIiwkHwofyva349tVu8QSDn",
            "rWuxi2uZmBEprBBpxLLFcKtXHA8bQkvJ",
            "8PV5LN1IGm5T0ZVIaakb218NvTEABNZz",
            "EMSmscnxwfrkKd1s3hOJ9bL4zqT1uud5",
            "YiLx0WGJrQAge2cVoAcCscDDVidbH4uE",
            "IjD0JHdQdDTwKusrbuiRO4NlFzbPotvg",
            "ADwiQogJGOS4X8dfIcidcfdT9fY2WpHC",
            "DyGaS7rfQsMX0E6TD9yORqx7kJgUYvNR",
            "MjOCkn11liCYZspTAhdULMEfWJGMHvpK",
            "wtGa41YcFMR5CBNr79lTfRAFi6Vhr6UF",
            "3AsMjnpTBawWv2AAPDxLjdxx4QYl9XXb",
            "sa2uMVNPiZLK52zzxlakCUXLaRNXddBz",
            "NYH9lkdbvXsiUFKwJtjSkQ1RzpHwWloK",
            "8AwzQeY5cudY8YUhwxm3UEP7Oos61RTY",
            "ADKWL3p5gloRYO3ptarTCbWUHo5JZi3j",
            "UAnMARj5x7hkh9kwBiNRfs5aYDsbHKpw",
            "Aes1rgTLMNnlCkb9s6bH7iT5CbZTdxUw",
            "eh1LYTOfgISrZ54B7JbldEpvqVur57tv",
            "u5A1wEYax1kD9HBeIjwyNAoubDreCsZ6",
            "mv70SFwpAOHRZt4dmuw5n2lAsM1lOrcx",
            "i4nIu0VZMuh5hLrUm9w2kqNxcfYY7Y3a",
            "UqfewK2qFZqfJ619RKkRiZeYtO21ngX1",
        ]
        expected = [
            False,
            True,
            True,
            True,
            True,
            False,
            True,
            True,
            False,
            False,
            True,
            False,
            True,
            False,
            False,
            False,
            False,
            False,
            True,
            True,
            False,
            False,
            True,
            True,
            False,
        ]

        for i in range(len(ids)):
            self.assertEqual(_should_sample(ids[i], 2), expected[i])

    def test_probability(self):
        ''' test that _should_sample approximates 1 in N sampling for random IDs '''
        tests_count = 50000
        error_margin = 0.05

        sample_rates = [1, 2, 10]

        for rate in sample_rates:
            sampled = n = 0

            while n < tests_count:
                n += 1
                if _should_sample(generate_trace_id(), rate):
                    sampled += 1

            expected = tests_count // rate

            acceptable_lower_bound = int(expected - (expected * error_margin))
            acceptable_upper_bound = int(expected + (expected * error_margin))

            self.assertLessEqual(sampled, acceptable_upper_bound)
            self.assertGreaterEqual(sampled, acceptable_lower_bound)

    def should_sample_always_returns_true_when_sample_rate_is_1(self):
        for _ in range(1000):
            self.assertTrue(_should_sample(binascii.b2a_hex(os.urandom(16)), 1))


class TestSynchronousTracer(unittest.TestCase):
    def test_trace_context_manager_exception(self):
        ''' ensure that span is sent even if an exception is
        raised inside the context manager '''
        m_client = Mock()
        tracer = SynchronousTracer(m_client)
        tracer.start_trace = Mock()
        mock_span = Mock()
        tracer.start_trace.return_value = mock_span
        tracer.finish_trace = Mock()
        try:
            with tracer('foo'):
                raise Exception('boom!')
        except Exception:
            pass
        tracer.finish_trace.assert_called_once_with(mock_span)

    def test_trace_context_manager_starts_span_if_trace_active(self):
        m_client = Mock()
        tracer = SynchronousTracer(m_client)
        tracer.start_span = Mock()
        tracer.start_trace = Mock()
        tracer.get_active_trace_id = Mock(return_value='asdf')
        mock_span = Mock()
        tracer.start_span.return_value = mock_span
        tracer.finish_span = Mock()

        with tracer('foo') as span:
            self.assertEqual(
                span, mock_span, 'tracer context manager should yield the span')

        tracer.start_span.assert_called_once_with(
            context={'name': 'foo'}, parent_id=None)
        tracer.start_trace.assert_not_called()
        tracer.finish_span.assert_called_once_with(mock_span)

    def test_trace_context_manager_passes_parent_id_if_supplied(self):
        ''' ensure parent_id gets passed to start_span if supplied '''
        m_client = Mock()
        tracer = SynchronousTracer(m_client)
        tracer.start_span = Mock()
        mock_span = Mock()
        tracer.start_span.return_value = mock_span
        tracer.finish_span = Mock()

        with tracer('foo', parent_id='zyxw'):
            pass

        tracer.start_span.assert_called_once_with(
            context={'name': 'foo'}, parent_id='zyxw')
        tracer.finish_span.assert_called_once_with(mock_span)

    def test_trace_context_manager_starts_trace_if_trace_id_supplied(self):
        ''' ensure trace_id and parent_id get passed to start_trace if supplied '''
        m_client = Mock()
        tracer = SynchronousTracer(m_client)
        tracer.start_trace = Mock()
        mock_span = Mock()
        tracer.start_trace.return_value = mock_span
        tracer.finish_span = Mock()

        with tracer('foo', trace_id='asdf', parent_id='zyxw'):
            pass

        tracer.start_trace.assert_called_once_with(
            context={'name': 'foo'}, trace_id='asdf', parent_span_id='zyxw')
        tracer.finish_span.assert_called_once_with(mock_span)

    def test_start_trace(self):
        m_client = Mock()
        tracer = SynchronousTracer(m_client)

        span = tracer.start_trace(context={'big': 'important_stuff'})
        self.assertIsInstance(span.event.start_time, datetime.datetime)
        # make sure our context got passed on to the event
        m_client.new_event.return_value.add.assert_has_calls([
            call(data={'big': 'important_stuff'}),
            call(data={
                'trace.trace_id': span.trace_id,
                'trace.parent_id': span.parent_id,
                'trace.span_id': span.id,
            }),
        ])
        self.assertEqual(tracer._trace.stack[0], span)
        # ensure we started a trace
        self.assertIsNotNone(tracer._trace)

    def test_start_span(self):
        m_client = Mock()
        tracer = SynchronousTracer(m_client)

        span = tracer.start_trace(context={'big': 'important_stuff'})
        # make sure this is the only event in the stack
        self.assertEqual(tracer._trace.stack[0], span)
        self.assertEqual(len(tracer._trace.stack), 1)

        span2 = tracer.start_span(context={'more': 'important_stuff'})
        # should still have the root span as the first item in the stack
        self.assertEqual(tracer._trace.stack[0], span)
        self.assertEqual(tracer._trace.stack[-1], span2)
        # should have the first span id as its parent
        # should share the same trace id
        self.assertEqual(span.trace_id, span2.trace_id)
        self.assertEqual(span.id, span2.parent_id)
        # trace id should match what the tracer has
        self.assertEqual(span.trace_id, tracer._trace.id)
        m_client.new_event.return_value.add.assert_has_calls([
            call(data={'more': 'important_stuff'}),
            call(data={
                'trace.trace_id': span2.trace_id,
                'trace.parent_id': span2.parent_id,
                'trace.span_id': span2.id,
            }),
        ])

    def test_start_span_returns_none_if_no_trace(self):
        m_client = Mock()
        tracer = SynchronousTracer(m_client)

        span = tracer.start_span(context={'more': 'important_stuff'})
        # should still have no active trace
        self.assertIsNone(span)
        self.assertIsNone(tracer._trace)

    def test_finish_trace(self):
        # implicitly tests finish_span
        m_client = Mock()
        # these values are used before sending
        m_client.new_event.return_value.start_time = datetime.datetime.now()
        m_client.new_event.return_value.sample_rate = 1
        tracer = SynchronousTracer(m_client)

        span = tracer.start_trace(context={'big': 'important_stuff'})
        self.assertEqual(tracer._trace.stack[0], span)

        tracer.finish_trace(span)
        # ensure the event is sent
        span.event.send_presampled.assert_called_once_with()
        # ensure that there is no current trace
        self.assertIsNone(tracer._trace)

    def test_start_trace_with_trace_id_set(self):
        m_client = Mock()
        tracer = SynchronousTracer(m_client)

        span = tracer.start_trace(trace_id='123456', parent_span_id='999999')
        self.assertEqual(span.trace_id, '123456')
        self.assertEqual(span.parent_id, '999999')
        self.assertEqual(tracer._trace.id, '123456')

        m_client.new_event.return_value.add.assert_has_calls([
            call(data={
                'trace.trace_id': span.trace_id,
                'trace.parent_id': span.parent_id,
                'trace.span_id': span.id,
            }),
        ])

    def test_add_trace_field_propagates(self):
        m_client = Mock()
        tracer = SynchronousTracer(m_client)

        span = tracer.start_trace(context={'big': 'important_stuff'})
        # make sure this is the only event in the stack
        self.assertEqual(tracer._trace.stack[0], span)
        self.assertEqual(len(tracer._trace.stack), 1)

        m_client.new_event.reset_mock()

        tracer.add_trace_field('another', 'important_thing')
        tracer.add_trace_field('wide', 'events_are_great')
        # ensure we don't crash if non-string added here
        tracer.add_trace_field(12345, 54321)
        # ensure fields prefixed with app. don't get another prefix
        tracer.add_trace_field('app.prefixes', 'there should only be 1')

        span2 = tracer.start_span(context={'more': 'important_stuff'})
        # should still have the root span as the first item in the stack
        self.assertEqual(tracer._trace.stack[0], span)
        self.assertEqual(tracer._trace.stack[-1], span2)
        # should have the first span id as its parent
        # should share the same trace id
        self.assertEqual(span.trace_id, span2.trace_id)
        self.assertEqual(span.id, span2.parent_id)
        # trace id should match what the tracer has
        self.assertEqual(span.trace_id, tracer._trace.id)
        m_client.new_event.assert_called_once_with(data={
            'app.another': 'important_thing',
            'app.wide': 'events_are_great',
            'app.12345': 54321,
            'app.prefixes': 'there should only be 1',
        })
        m_client.new_event.return_value.add.assert_has_calls([
            call(data={'more': 'important_stuff'}),
            call(data={
                'trace.trace_id': span2.trace_id,
                'trace.parent_id': span2.parent_id,
                'trace.span_id': span2.id,
            }),
        ])

        m_client.new_event.reset_mock()
        m_client.new_event.return_value.fields.return_value = {}
        # swap out some trace fields
        tracer.add_trace_field('more', 'data!')
        tracer.remove_trace_field('another')

        span3 = tracer.start_span(context={'more': 'important_stuff'})
        self.assertEqual(tracer._trace.stack[0], span)
        self.assertEqual(tracer._trace.stack[1], span2)
        self.assertEqual(tracer._trace.stack[-1], span3)
        # should have the second span id as its parent
        # should share the same trace id
        self.assertEqual(span.trace_id, span3.trace_id)
        self.assertEqual(span2.id, span3.parent_id)
        m_client.new_event.assert_called_once_with(data={
            'app.wide': 'events_are_great',
            'app.more': 'data!',
            'app.12345': 54321,
            'app.prefixes': 'there should only be 1',
        })

    def test_add_rollup_field_propagates(self):
        m_client = Mock()
        tracer = SynchronousTracer(m_client)
        tracer._run_hooks_and_send = Mock()

        span1 = tracer.start_trace(context={'name': 'root'})
        event1 = m_client.new_event.return_value

        span2 = tracer.start_span(context={'name': 'middle'})
        event2 = m_client.new_event.return_value

        span3 = tracer.start_span(context={'name': 'inner1'})
        event3 = m_client.new_event.return_value

        tracer.add_rollup_field('database_ms', 17)
        tracer.add_rollup_field('calories', 180)
        tracer.add_rollup_field('database_ms', 23.1)

        event3.add_field.reset_mock()
        tracer.finish_span(span3)
        event3.add_field.assert_has_calls([
            call('database_ms', 17.0 + 23.1),
            call('calories', 180.0),
            call('duration_ms', ANY),
        ], any_order=True)

        span4 = tracer.start_span(context={'name': 'inner2'})
        event4 = m_client.new_event.return_value

        tracer.add_rollup_field('calories', 120)

        event4.add_field.reset_mock()
        tracer.finish_span(span4)
        event4.add_field.assert_has_calls([
            call('calories', 120.0),
            call('duration_ms', ANY),
        ], any_order=True)

        event2.add_field.reset_mock()
        tracer.finish_span(span2)
        event2.add_field.assert_has_calls([
            # This intermediate span doesn't get any rollup fields.
            call('duration_ms', ANY),
        ], any_order=True)

        event1.add_field.reset_mock()
        tracer.finish_span(span1)
        event1.add_field.assert_has_calls([
            call('rollup.database_ms', 17.0 + 23.1),
            call('rollup.calories', 180.0 + 120.0),
            call('duration_ms', ANY),
        ], any_order=True)

    def test_get_active_span(self):
        m_client = Mock()
        tracer = SynchronousTracer(m_client)
        span = tracer.start_trace()
        self.assertEqual(tracer.get_active_span().id, span.id)

    def test_run_hooks_and_send_no_hooks(self):
        ''' ensure send works when no hooks defined '''
        m_client = Mock()
        tracer = SynchronousTracer(m_client)
        m_span = Mock()

        with patch('beeline.trace._should_sample') as m_sample_fn:
            m_sample_fn.return_value = True
            tracer._run_hooks_and_send(m_span)

        # no hooks - trace's _should_sample is rigged to always return True, so we
        # always call send_presampled
        # send should never be called because at a minimum we always do deterministic
        # sampling
        m_span.event.send.assert_not_called()  # pylint: disable=no-member
        m_span.event.send_presampled.assert_called_once_with()  # pylint: disable=no-member

    def test_run_hooks_and_send_sampler(self):
        ''' ensure send works with a sampler hook defined '''
        m_client = Mock()
        tracer = SynchronousTracer(m_client)
        m_span = Mock()

        def _sampler_drop_all(fields):
            return False, 0

        tracer.register_hooks(sampler=_sampler_drop_all)

        with patch('beeline.trace._should_sample') as m_sample_fn:
            m_sample_fn.return_value = True
            tracer._run_hooks_and_send(m_span)

        # sampler ensures we drop everything
        m_span.event.send.assert_not_called()  # pylint: disable=no-member
        m_span.event.send_presampled.assert_not_called()  # pylint: disable=no-member

        def _sampler_drop_none(fields):
            return True, 100

        tracer.register_hooks(sampler=_sampler_drop_none)
        m_span.reset_mock()

        with patch('beeline.trace._should_sample') as m_sample_fn:
            m_sample_fn.return_value = True
            tracer._run_hooks_and_send(m_span)

        # sampler drops nothing, _should_sample rigged to always return true so
        # we always call send_presampled
        m_span.event.send.assert_not_called()  # pylint: disable=no-member
        m_span.event.send_presampled.assert_called_once_with()  # pylint: disable=no-member
        # ensure event is updated with new sample rate
        self.assertEqual(m_span.event.sample_rate, 100)

    def test_run_hooks_and_send_presend_hook(self):
        ''' ensure send works when presend hook is defined '''
        m_client = Mock()
        tracer = SynchronousTracer(m_client)
        m_span = Mock()

        def _presend_hook(fields):
            fields["thing i want"] = "put it there"
            del fields["thing i don't want"]

        m_span = Mock()
        m_span.event.fields.return_value = {
            "thing i don't want": "get it out of here",
            "happy data": "so happy",
        }

        tracer.register_hooks(presend=_presend_hook)

        with patch('beeline.trace._should_sample') as m_sample_fn:
            m_sample_fn.return_value = True
            tracer._run_hooks_and_send(m_span)

        m_span.event.send_presampled.assert_called_once_with()  # pylint: disable=no-member
        self.assertDictEqual(
            m_span.event.fields(),
            {
                "thing i want": "put it there",
                "happy data": "so happy",
            },
        )

    def test_run_hooks_and_send_adds_trace_fields(self):
        ''' ensure trace fields are propagated backwards '''
        m_client = Mock()
        tracer = SynchronousTracer(m_client)
        tracer.start_trace()
        m_span = Mock()
        m_span.event = Event()
        m_span.event.start_time = datetime.datetime.now()
        # set an existing trace field
        m_span.event.add_field('app.a', 1)
        m_span.rollup_fields = defaultdict(float)

        with patch('beeline.trace._should_sample') as m_sample_fn:
            m_sample_fn.return_value = True
            # add some trace fields
            tracer.add_trace_field('a', 0)
            tracer.add_trace_field('b', 2)
            tracer.add_trace_field('c', 3)
            tracer.finish_span(m_span)

        # Check that the new, unique fields were successfully added
        self.assertIn("app.b", m_span.event.fields())
        self.assertIn("app.c", m_span.event.fields())

        # Check that a was not overwritten with the new value of 0.
        self.assertEqual(m_span.event.fields()["app.a"], 1)

    def test_trace_context_manager_does_not_crash_if_span_is_none(self):
        m_client = Mock()
        tracer = SynchronousTracer(m_client)
        tracer.start_span = Mock()
        tracer.start_span.return_value = None
        tracer.finish_span = Mock()

        with tracer('foo'):
            pass

        tracer.start_span.assert_called_once_with(
            context={'name': 'foo'}, parent_id=None)

    def test_trace_with_custom_dataset(self):
        dataset = 'flibble'
        m_client = Mock()
        tracer = SynchronousTracer(m_client)
        tracer.start_trace(dataset=dataset)
        self.assertEqual(tracer._trace.dataset, dataset)

    def test_custom_dataset_propagates_to_event(self):
        dataset = 'flibble'
        m_client = Mock()
        tracer = SynchronousTracer(m_client)
        tracer._run_hooks_and_send = Mock()
        span = tracer.start_trace(dataset=dataset)
        event = m_client.new_event.return_value
        tracer.finish_span(span)
        self.assertEqual(event.dataset, dataset)


class TestTraceContext(unittest.TestCase):
    def test_marshal_trace_context(self):
        trace_id = "123456"
        parent_id = "654321"
        trace_fields = {"i": "like", "to": "trace"}

        trace_context = marshal_trace_context(
            trace_id, parent_id, trace_fields)

        trace_id_u, parent_id_u, trace_fields_u = unmarshal_trace_context(
            trace_context)
        self.assertEqual(trace_id_u, trace_id,
                         "unmarshaled trace id should match original")
        self.assertEqual(parent_id_u, parent_id,
                         "unmarshaled parent id should match original")
        self.assertDictEqual(trace_fields_u, trace_fields,
                             "unmarshaled trace fields should match original")

    def test_marshal_trace_context_empty_context(self):
        trace_id = "123456"
        parent_id = "654321"
        trace_context = "{};trace_id={},parent_id={}".format(
            1, trace_id, parent_id)

        trace_id_u, parent_id_u, trace_fields_u = unmarshal_trace_context(
            trace_context)
        self.assertEqual(trace_id_u, trace_id,
                         "unmarshaled trace id should match original")
        self.assertEqual(parent_id_u, parent_id,
                         "unmarshaled parent id should match original")
        self.assertDictEqual(trace_fields_u, {},
                             "unmarshaled trace fields should match original")

    def test_marshal_trace_context_dataset_included(self):
        """ ensures unmarshalling still works if there's a dataset context field """
        trace_id = "123456"
        parent_id = "654321"
        dataset_id = "foo"
        trace_fields = {"i": "like", "to": "trace"}
        fields_string = base64.b64encode(
            json.dumps(trace_fields).encode()).decode()
        trace_context = "{};trace_id={},parent_id={},dataset_id={},context={}".format(
            1, trace_id, parent_id, dataset_id, fields_string
        )

        trace_id_u, parent_id_u, trace_fields_u = unmarshal_trace_context(
            trace_context)
        self.assertEqual(trace_id_u, trace_id,
                         "unmarshaled trace id should match original")
        self.assertEqual(parent_id_u, parent_id,
                         "unmarshaled parent id should match original")
        self.assertDictEqual(trace_fields_u, trace_fields,
                             "unmarshaled trace fields should match original")


class TestSpan(unittest.TestCase):
    def test_span_context(self):
        ev = Event()
        span = Span('', '', '', ev)
        span.add_context_field("some", "value")
        span.add_context({"another": "value"})
        self.assertDictEqual({
            "some": "value",
            "another": "value"
        }, ev.fields())
        span.remove_context_field("another")
        self.assertDictEqual({
            "some": "value",
        }, ev.fields())


class TestPropagationHooks(unittest.TestCase):
    def test_propagate_and_start_trace_uses_honeycomb_header(self):
        beeline.propagation.propagate_dataset = True
        # FIXME: Test basics, including error handling and custom hooks

        # implicitly tests finish_span
        m_client = Mock()
        # these values are used before sending
        m_client.new_event.return_value.start_time = datetime.datetime.now()
        m_client.new_event.return_value.sample_rate = 1
        tracer = SynchronousTracer(m_client)

        header_value = '1;dataset=flibble,trace_id=bloop,parent_id=scoop,context=e30K'
        req = DictRequest({
            # case shouldn't matter
            'X-HoNEyComb-TrACE': header_value,
        })

        span = tracer.propagate_and_start_trace(
            context={'big': 'important_stuff'}, request=req)
        self.assertEqual(tracer._trace.stack[0], span)
        self.assertEqual(span.trace_id, "bloop")
        self.assertEqual(span.parent_id, "scoop")

        tracer.finish_trace(span)
        self.assertEqual(span.event.dataset, 'flibble')

        # ensure the event is sent
        span.event.send_presampled.assert_called_once_with()
        # ensure that there is no current trace
        self.assertIsNone(tracer._trace)

    def test_propagate_and_start_trace_uses_honeycomb_header_with_dataset_propagation_disabled(self):
        beeline.propagation.propagate_dataset = False
        # FIXME: Test basics, including error handling and custom hooks

        # implicitly tests finish_span
        m_client = Mock()
        # these values are used before sending
        m_client.new_event.return_value.start_time = datetime.datetime.now()
        m_client.new_event.return_value.sample_rate = 1
        tracer = SynchronousTracer(m_client)

        header_value = '1;dataset=flibble,trace_id=bloop,parent_id=scoop,context=e30K'
        req = DictRequest({
            # case shouldn't matter
            'X-HoNEyComb-TrACE': header_value,
        })

        span = tracer.propagate_and_start_trace(
            context={'big': 'important_stuff'}, request=req)
        self.assertEqual(tracer._trace.stack[0], span)
        self.assertEqual(span.trace_id, "bloop")
        self.assertEqual(span.parent_id, "scoop")

        tracer.finish_trace(span)

        # testing the absence of dataset
        self.assertNotEqual(getattr(span.event, 'dataset'), "flibble")
        self.assertNotIsInstance(span.event.dataset, str)

        # ensure the event is sent
        span.event.send_presampled.assert_called_once_with()
        # ensure that there is no current trace
        self.assertIsNone(tracer._trace)

    def test_propagate_and_start_trace_uses_w3c_header_as_fallback(self):
        # implicitly tests finish_span
        m_client = Mock()
        # these values are used before sending
        m_client.new_event.return_value.start_time = datetime.datetime.now()
        m_client.new_event.return_value.sample_rate = 1
        tracer = SynchronousTracer(m_client)

        header_value = '00-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-00'
        req = DictRequest({
            'traceparent': header_value,
        })

        span = tracer.propagate_and_start_trace(
            context={'big': 'important_stuff'}, request=req)
        self.assertEqual(tracer._trace.stack[0], span)
        self.assertEqual(span.trace_id, "0af7651916cd43dd8448eb211c80319c")
        self.assertEqual(span.parent_id, "b7ad6b7169203331")

        tracer.finish_trace(span)

        # ensure the event is sent
        span.event.send_presampled.assert_called_once_with()
        # ensure that there is no current trace
        self.assertIsNone(tracer._trace)

    def test_propagate_and_start_trace_uses_honeycomb_header_when_w3c_also_present(self):
        # implicitly tests finish_span
        m_client = Mock()
        # these values are used before sending
        m_client.new_event.return_value.start_time = datetime.datetime.now()
        m_client.new_event.return_value.sample_rate = 1
        tracer = SynchronousTracer(m_client)

        w3c_value = '00-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-00'
        honeycomb_value = '1;dataset=flibble,trace_id=bloop,parent_id=scoop,context=e30K'
        req = DictRequest({
            'traceparent': w3c_value,
            'x-honeycomb-trace': honeycomb_value
        })

        span = tracer.propagate_and_start_trace(
            context={'big': 'important_stuff'}, request=req)
        self.assertEqual(tracer._trace.stack[0], span)
        self.assertEqual(span.trace_id, "bloop")
        self.assertEqual(span.parent_id, "scoop")

        tracer.finish_trace(span)
        self.assertEqual(span.event.dataset, 'flibble')

        # ensure the event is sent
        span.event.send_presampled.assert_called_once_with()
        # ensure that there is no current trace
        self.assertIsNone(tracer._trace)

    def test_error_handling(self):
        class TestException(Exception):
            pass

        def error_parser(propagation_context):
            raise TestException()

        # implicitly tests finish_span
        m_client = Mock()
        # these values are used before sending
        m_client.new_event.return_value.start_time = datetime.datetime.now()
        m_client.new_event.return_value.sample_rate = 99
        tracer = SynchronousTracer(m_client)
        tracer.register_hooks(http_trace_parser=error_parser)

        header_value = '1;trace_id=bloop,parent_id=scoop,context=e30K'
        req = DictRequest({
            # case shouldn't matter
            'X-HoNEyComb-TrACE': header_value,
        })

        span = tracer.propagate_and_start_trace(
            context={'big': 'important_stuff'}, request=req)

        # Verify that our new span is at the top of the stack
        self.assertEqual(tracer._trace.stack[0], span)
        # Verify that we got the event in the mock.
        self.assertEqual(span.event.sample_rate, 99)



from beeline.propagation import honeycomb, w3c


def http_trace_parser_hook(request):
    """
    Retrieves the propagation context out of the request. Uses the honeycomb header, with W3C header as fallback.
    """
    honeycomb_header_value = honeycomb.http_trace_parser_hook(request)
    w3c_header_value = w3c.http_trace_parser_hook(request)
    if honeycomb_header_value:
        return honeycomb_header_value
    else:
        return w3c_header_value


def http_trace_propagation_hook(propagation_context):
    """
    Given a propagation context, returns a dictionary of key value pairs that should be
    added to outbound requests (usually HTTP headers). Uses the honeycomb format.
    """
    return honeycomb.http_trace_propagation_hook(propagation_context)



''' patches base tornado classes to add honeycomb instrumentation '''

from wrapt import wrap_function_wrapper
import beeline
import tornado
assert tornado  # for pyflakes


def log_request(_log_request, instance, args, kwargs):
    try:
        # expecting signature `log_request(self, handler)`
        if len(args) == 1:
            handler = args[0]
            beeline.send_now({
                "duration_ms": handler.request.request_time() * 1000.0,
                "request.method": handler.request.method,
                "request.remote_addr": handler.request.remote_ip,
                "request.path": handler.request.uri,
                "request.query": handler.request.query,
                "request.host": handler.request.headers.get('Host'),
                "response.status_code": handler.get_status(),
            })
    except Exception:
        pass
    finally:
        _log_request(*args, **kwargs)


def log_exception(_log_exception, instance, args, kwargs):
    try:
        # expecting signature `log_exception(self, typ, value, tb)``
        if len(args) == 3:
            value = args[2]
            beeline.send_now({
                "request.method": instance.request.method,
                "request.path": instance.request.uri,
                "request.remote_addr": instance.request.remote_ip,
                "request.query": instance.request.query,
                "request.host": instance.request.get('Host'),
                "request.error": type(value).__name__,
                "request.error_detail": beeline.internal.stringify_exception(value),
            })
    except Exception:
        pass
    finally:
        _log_exception(*args, **kwargs)


wrap_function_wrapper('tornado.web', 'Application.log_request', log_request)
wrap_function_wrapper(
    'tornado.web', 'RequestHandler.log_exception', log_exception)



import unittest
try:
    # The async functionality uses the contextvars module, added in
    # Python 3.7
    import contextvars
except ImportError:
    contextvars = None
if not contextvars:
    raise unittest.SkipTest("No contextvars failed. Skipping test_async")

import asyncio
import concurrent.futures
import datetime
import time
import sys

import beeline
import beeline.aiotrace
import beeline.trace


def async_test(fn):
    """Decorator for making async methods usable as tests.

    This decorator also runs the async_setup method before each test,
    if it exists.

    """

    def wrapper(self, *args, **kwargs):
        async def sequence():
            if hasattr(self, "async_setup"):
                await self.async_setup()
            return await fn(self, *args, **kwargs)

        return asyncio.run(sequence(*args, **kwargs))  # pylint: disable=no-member

    return wrapper


def span_data(span):
    """Utility for converting Span objects to dicts."""
    name = span.event.fields().get("name")
    start = span.event.start_time
    duration = datetime.timedelta(
        milliseconds=span.event.fields()["duration_ms"]
    )
    end = start + duration
    return {
        "name": name,
        "start": start,
        "end": end,
        "trace_id": span.trace_id,
        "span": span,
    }


class TestTracerImplChoice(unittest.TestCase):
    def test_synchronous_tracer_should_be_used_by_default(self):
        """Verify that the SynchronousTracer implementation is chosen when a
        Beeline object is initialised outside of an asyncio loop.

        """
        _beeline = beeline.Beeline()
        self.assertIsInstance(
            _beeline.tracer_impl, beeline.trace.SynchronousTracer
        )

    @async_test
    async def test_asyncio_tracer_should_be_used_in_async_code(self):
        """Verify that the AsyncioTracer implementation is chosen when a
        Beeline object is initialised while running inside an asyncio
        loop.

        """
        _beeline = beeline.Beeline()
        self.assertIsInstance(
            _beeline.tracer_impl, beeline.aiotrace.AsyncioTracer
        )


class TestAsynchronousTracer(unittest.TestCase):
    async def async_setup(self):
        self.finished_spans = []

        def add_span(span):
            self.finished_spans.append(span_data(span))

        self.beeline = beeline.Beeline()
        self.tracer = self.beeline.tracer_impl
        self.tracer._run_hooks_and_send = add_span

    @async_test
    async def test_tracing_in_new_tasks_should_work(self):
        """Test that basic AsyncioTracer functionality is present."""
        trace = self.tracer.start_trace()
        self.tracer.finish_trace(trace)

        self.assertEqual(len(self.finished_spans), 1)

    @async_test
    async def test_traces_started_in_different_tasks_should_be_independent(self):
        """Fork off two tasks, each calling start_trace.

        The traces run simultaneously. This is expected to produce two
        independent traces without raising any exceptions.

        """
        async def task0():
            trace0 = self.tracer.start_trace(context={"name": "task0"})
            await asyncio.sleep(0.2)
            self.tracer.finish_trace(trace0)

        async def task1():
            await asyncio.sleep(0.1)
            trace1 = self.tracer.start_trace(context={"name": "task1"})
            await asyncio.sleep(0.2)
            self.tracer.finish_trace(trace1)

        await asyncio.gather(task0(), task1())

        self.assertEqual(len(self.finished_spans), 2)
        task0_span, task1_span = self.finished_spans  # pylint: disable=unbalanced-tuple-unpacking

        # Check that the spans finished in the expected order.
        self.assertEqual(task0_span["name"], "task0")
        self.assertEqual(task1_span["name"], "task1")
        self.assertLess(task0_span["end"], task1_span["end"])

        # Check that the task0 started before task1
        self.assertLess(task0_span["start"], task1_span["start"])

        # Check that the task1 span started during the task0 span
        self.assertLess(task1_span["start"], task0_span["end"])

        # Check that the task spans are both root spans
        self.assertTrue(task0_span["span"].is_root())
        self.assertTrue(task1_span["span"].is_root())

    @async_test
    async def test_new_tasks_should_trace_in_parallel(self):
        """Fork off two tasks after starting a trace.

        Both tasks record a span, overlapping with each other. Both
        spans are expected to have the root span as their parent.

        """

        trace = self.tracer.start_trace(context={"name": "root"})

        async def task0():
            span0 = self.tracer.start_span(context={"name": "task0"})
            await asyncio.sleep(0.2)
            self.tracer.finish_span(span0)

        async def task1():
            await asyncio.sleep(0.1)
            span1 = self.tracer.start_span(context={"name": "task1"})
            await asyncio.sleep(0.2)
            self.tracer.finish_span(span1)

        await asyncio.gather(task0(), task1())

        self.tracer.finish_trace(trace)

        self.assertEqual(len(self.finished_spans), 3)
        task0_span, task1_span, root_span = self.finished_spans  # pylint: disable=unbalanced-tuple-unpacking

        # Check that the spans finished in the expected order, with
        # the root span last.
        self.assertEqual(task0_span["name"], "task0")
        self.assertEqual(task1_span["name"], "task1")
        self.assertLess(task0_span["end"], task1_span["end"])
        self.assertEqual(root_span["name"], "root")
        self.assertLessEqual(task1_span["end"], root_span["end"])

        # Check that the root span was started before the others.
        self.assertLess(root_span["start"], task0_span["start"])
        self.assertLess(root_span["start"], task1_span["start"])

        # Check that the task0 started before task1
        self.assertLess(task0_span["start"], task1_span["start"])

        # Check that the task1 span started during the task0 span
        self.assertLess(task1_span["start"], task0_span["end"])

        # Check that the task spans are both children of the root span
        self.assertEqual(root_span["span"].id, task0_span["span"].parent_id)
        self.assertEqual(root_span["span"].id, task1_span["span"].parent_id)

    @async_test
    async def test_traced_decorators(self):
        """Fork off two tasks after starting a trace.

        This is the same as test_new_tasks_should_trace_in_parallel,
        except it uses the traced decorator to record the sub-spans.

        """
        trace = self.tracer.start_trace(context={"name": "root"})

        @self.beeline.traced("task0")
        async def task0():
            await asyncio.sleep(0.2)

        async def task1():
            await asyncio.sleep(0.1)

            @self.beeline.traced("task1")
            async def decorated_fn():
                await asyncio.sleep(0.2)

            await decorated_fn()

        await asyncio.gather(task0(), task1())

        self.tracer.finish_trace(trace)

        self.assertEqual(len(self.finished_spans), 3)

        task0_span, task1_span, root_span = self.finished_spans  # pylint: disable=unbalanced-tuple-unpacking

        # Check that the spans finished in the expected order, with
        # the root span last.
        self.assertEqual(task0_span["name"], "task0")
        self.assertEqual(task1_span["name"], "task1")
        self.assertLess(task0_span["end"], task1_span["end"])
        self.assertEqual(root_span["name"], "root")
        self.assertLessEqual(task1_span["end"], root_span["end"])

        # Check that the root span was started before the others.
        self.assertLess(root_span["start"], task0_span["start"])
        self.assertLess(root_span["start"], task1_span["start"])

        # Check that the task0 started before task1
        self.assertLess(task0_span["start"], task1_span["start"])

        # Check that the task1 span started during the task0 span
        self.assertLess(task1_span["start"], task0_span["end"])

        # Check that the task spans are both children of the root span
        self.assertEqual(root_span["span"].id, task0_span["span"].parent_id)
        self.assertEqual(root_span["span"].id, task1_span["span"].parent_id)

    @async_test
    async def test_traceless_spans_in_other_tasks_should_be_ignored(self):
        """Start a span without first starting a trace in the same task.

        This span is started while there is a trace started in another
        task. This span should be independent from that trace, and is
        expected to be ignored.

        """
        async def task0():
            await asyncio.sleep(0.2)
            trace = self.tracer.start_trace(context={"name": "task0"})
            await asyncio.sleep(0.2)
            self.tracer.finish_trace(trace)

        async def task1():
            await asyncio.sleep(0.1)
            span = self.tracer.start_span(context={"name": "task1"})
            await asyncio.sleep(0.2)
            self.tracer.finish_span(span)

        await asyncio.gather(task0(), task1())

        self.assertEqual(len(self.finished_spans), 1)
        task0_span = self.finished_spans[0]

        # Check that only the trace produced a span.
        self.assertEqual(task0_span["name"], "task0")

    @async_test
    async def test_untraced_async_functions_should_work(self):
        """Call functions with the untraced decorator from within a trace.

        A trace is started and untraced async functions are called.
        They start spans, but the spans are not expected to be
        recorded since they should be considered started outside of
        any trace.

        """

        calls = set()

        trace = self.tracer.start_trace(context={"name": "root"})

        @beeline.untraced
        async def fn0():
            span0 = self.tracer.start_span(context={"name": "fn0"})
            await asyncio.sleep(0.2)
            self.tracer.finish_span(span0)
            calls.add("fn0")

        @beeline.untraced
        async def fn1():
            await asyncio.sleep(0.1)
            span1 = self.tracer.start_span(context={"name": "fn1"})
            await asyncio.sleep(0.2)
            self.tracer.finish_span(span1)
            calls.add("fn1")

        # Use the decorated function as a plain coroutine
        await fn0()

        # Use the decorated function as a task
        task = asyncio.create_task(fn1())  # pylint: disable=no-member
        await task

        self.tracer.finish_trace(trace)

        self.assertEqual(len(self.finished_spans), 1)
        root_span = self.finished_spans[0]

        self.assertEqual(root_span["name"], "root")

        # Verify that the untraced functions were actually called
        self.assertTrue("fn0" in calls)
        self.assertTrue("fn1" in calls)

    @async_test
    async def test_untraced_synchronous_functions_should_work(self):
        """Call synchronous functions with the untraced decorator.

        A trace is started and untraced synchronous functions are
        called. They start spans, but the spans are not expected to be
        recorded since they should be considered started outside of
        any trace.

        """

        calls = set()

        trace = self.tracer.start_trace(context={"name": "root"})

        @beeline.untraced
        def fn0():
            span0 = self.tracer.start_span(context={"name": "fn0"})
            self.tracer.finish_span(span0)
            calls.add("fn0")

        @beeline.untraced
        def fn1():
            async def task1():
                await asyncio.sleep(0.1)
                span1 = self.tracer.start_span(context={"name": "fn1"})
                await asyncio.sleep(0.2)
                self.tracer.finish_span(span1)
                calls.add("fn1")

            return asyncio.create_task(task1())  # pylint: disable=no-member

        # Call one synchronous function
        fn0()
        # Spawn a task within another synchronous function
        task = fn1()

        await task

        self.tracer.finish_trace(trace)

        self.assertEqual(len(self.finished_spans), 1)
        root_span = self.finished_spans[0]

        self.assertEqual(root_span["name"], "root")

        # Verify that the untraced functions were actually called
        self.assertTrue("fn0" in calls)
        self.assertTrue("fn1" in calls)

    @async_test
    async def test_traced_thread_should_work_with_async_tracer(self):
        """Run traced code in threads from within async code.

        A trace is started and two functions are run in threads via a
        ThreadPoolExecutor. Both functions start spans, but only one
        of them is decorated with the traced_thread decorator. Only
        the span from the decorated function is expected to show up in
        the trace.

        """
        loop = asyncio.get_running_loop()  # pylint: disable=no-member
        executor = concurrent.futures.ThreadPoolExecutor(max_workers=2)
        calls = set()

        trace = self.tracer.start_trace(context={"name": "root"})

        @self.beeline.traced_thread
        def traced_worker():
            span = self.tracer.start_span(context={"name": "traced_worker"})
            time.sleep(0.2)
            self.tracer.finish_span(span)

        def untraced_worker():
            span = self.tracer.start_span(context={"name": "untraced_worker"})
            time.sleep(0.2)
            self.tracer.finish_span(span)
            calls.add("untraced_worker")

        future0 = loop.run_in_executor(executor, traced_worker)
        future1 = loop.run_in_executor(executor, untraced_worker)

        await asyncio.gather(future0, future1)

        self.tracer.finish_trace(trace)

        self.assertEqual(len(self.finished_spans), 2)

        worker_span, root_span = self.finished_spans  # pylint: disable=unbalanced-tuple-unpacking

        # Check that the spans finished in the expected order, with
        # the root span last.
        self.assertEqual(worker_span["name"], "traced_worker")
        self.assertEqual(root_span["name"], "root")
        self.assertLessEqual(worker_span["end"], root_span["end"])

        # Check that the root span was started before the worker span.
        self.assertLess(root_span["start"], worker_span["start"])

        # Check that the worker span is a child of the root span.
        self.assertEqual(root_span["span"].id, worker_span["span"].parent_id)

        # Verify that the untraced function was actually called
        self.assertTrue("untraced_worker" in calls)



import unittest
import beeline.propagation
from beeline.propagation import DictRequest, PropagationContext
import beeline.propagation.honeycomb as hc

header_value = '1;trace_id=bloop,parent_id=scoop,context=e30K'


class TestMarshalUnmarshal(unittest.TestCase):
    def test_roundtrip(self):
        '''Verify that we can successfully roundtrip (marshal and unmarshal)'''
        trace_id = "bloop"
        parent_id = "scoop"
        trace_fields = {"key": "value"}
        pc = PropagationContext(trace_id, parent_id, trace_fields)
        header = hc.marshal_propagation_context(pc)
        new_trace_id, new_parent_id, new_trace_fields = hc.unmarshal_propagation_context(
            header)
        self.assertEqual(trace_id, new_trace_id)
        self.assertEqual(parent_id, new_parent_id)
        self.assertEqual(trace_fields, new_trace_fields)

    def test_roundtrip_with_dataset(self):
        '''Verify that we can successfully roundtrip (marshal and unmarshal)'''
        beeline.propagation.propagate_dataset = True
        dataset = "blorp blorp"
        trace_id = "bloop"
        parent_id = "scoop"
        trace_fields = {"key": "value"}
        pc = PropagationContext(trace_id, parent_id, trace_fields, dataset)
        header = hc.marshal_propagation_context(pc)
        new_trace_id, new_parent_id, new_trace_fields, new_dataset = hc.unmarshal_propagation_context_with_dataset(
            header)
        self.assertEqual(dataset, new_dataset)
        self.assertEqual(trace_id, new_trace_id)
        self.assertEqual(parent_id, new_parent_id)
        self.assertEqual(trace_fields, new_trace_fields)

    def test_roundtrip_with_dataset_propagation_disabled(self):
        '''Verify that we can successfully roundtrip (marshal and unmarshal) without dataset propagation'''
        beeline.propagation.propagate_dataset = False
        dataset = "blorp blorp"
        trace_id = "bloop"
        parent_id = "scoop"
        trace_fields = {"key": "value"}
        pc = PropagationContext(trace_id, parent_id, trace_fields, dataset)
        header = hc.marshal_propagation_context(pc)
        new_trace_id, new_parent_id, new_trace_fields, new_dataset = hc.unmarshal_propagation_context_with_dataset(
            header)
        self.assertIsNone(new_dataset)
        self.assertEqual(trace_id, new_trace_id)
        self.assertEqual(parent_id, new_parent_id)
        self.assertEqual(trace_fields, new_trace_fields)


class TestHoneycombHTTPTraceParserHook(unittest.TestCase):
    def test_has_header(self):
        '''Test that the hook properly parses honeycomb trace headers'''
        req = DictRequest({
            # case shouldn't matter
            'X-HoNEyComb-TrACE': header_value,
        })
        pc = hc.http_trace_parser_hook(req)
        self.assertEqual(pc.trace_id, "bloop")
        self.assertEqual(pc.parent_id, "scoop")
        # FIXME: We should have a legitimate header with trace_field and dataset_id set

    def test_no_header(self):
        req = DictRequest({})
        pc = hc.http_trace_parser_hook(req)
        self.assertIsNone(pc)


class TestHoneycombHTTPTracePropagationHook(unittest.TestCase):
    def test_generates_correct_header(self):
        beeline.propagation.propagate_dataset = True
        dataset = "blorp blorp"
        trace_id = "bloop"
        parent_id = "scoop"
        trace_fields = {"key": "value"}
        pc = PropagationContext(
            trace_id, parent_id, trace_fields, dataset)
        headers = hc.http_trace_propagation_hook(pc)
        self.assertIn('X-Honeycomb-Trace', headers)
        self.assertEqual(headers['X-Honeycomb-Trace'],
                         "1;dataset=blorp%20blorp,trace_id=bloop,parent_id=scoop,context=eyJrZXkiOiAidmFsdWUifQ==")

    def test_generates_correct_header_with_dataset_propagation_disabled(self):
        beeline.propagation.propagate_dataset = False
        dataset = "blorp blorp"
        trace_id = "bloop"
        parent_id = "scoop"
        trace_fields = {"key": "value"}
        pc = PropagationContext(
            trace_id, parent_id, trace_fields, dataset)
        headers = hc.http_trace_propagation_hook(pc)
        self.assertIn('X-Honeycomb-Trace', headers)
        self.assertEqual(headers['X-Honeycomb-Trace'],
                         "1;trace_id=bloop,parent_id=scoop,context=eyJrZXkiOiAidmFsdWUifQ==")



import contextlib
import datetime
import beeline
from beeline.propagation import Request
from django.db import connections


class DjangoRequest(Request):
    def __init__(self, request):
        self._request = request
        self._META = request.META

        # only write log if beeline has been initalised
        if beeline.get_beeline():
            beeline.get_beeline().log(request.META)

    def header(self, key):
        lookup_key = "HTTP_" + key.upper().replace('-', '_')
        return self._request.META.get(lookup_key)

    def method(self):
        return self._request.method

    def scheme(self):
        return self._request.scheme

    def host(self):
        return self._request.get_host()

    def path(self):
        return self._request.path

    def query(self):
        return self._request.META.get('QUERY_STRING')

    def middleware_request(self):
        return self._request


class HoneyDBWrapper(object):

    def __call__(self, execute, sql, params, many, context):
        # if beeline has not been initialised, just execute query
        if not beeline.get_beeline():
            return execute(sql, params, many, context)

        vendor = context['connection'].vendor
        trace_name = "django_%s_query" % vendor

        with beeline.tracer(trace_name):
            beeline.add_context({
                "type": "db",
                "db.query": sql,
                "db.query_args": params,
            })
            beeline.add_rollup_field("db.call_count", 1)

            try:
                db_call_start = datetime.datetime.now()
                result = execute(sql, params, many, context)
                db_call_diff = datetime.datetime.now() - db_call_start
                duration = db_call_diff.total_seconds() * 1000
                beeline.add_context_field("db.duration", duration)
                beeline.add_rollup_field("db.total_duration", duration)
            except Exception as e:
                beeline.add_context_field("db.error", str(type(e)))
                beeline.add_context_field(
                    "db.error_detail", beeline.internal.stringify_exception(e))
                raise
            else:
                return result
            finally:
                if vendor in ('postgresql', 'mysql'):
                    beeline.add_context({
                        "db.last_insert_id": context['cursor'].cursor.lastrowid,
                        "db.rows_affected": context['cursor'].cursor.rowcount,
                    })


class HoneyMiddlewareBase(object):
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        response = self.create_http_event(request)
        return response

    def get_context_from_request(self, request):
        trace_name = "django_http_%s" % request.method.lower()
        return {
            "name": trace_name,
            "type": "http_server",
            "request.host": request.get_host(),
            "request.method": request.method,
            "request.path": request.path,
            "request.remote_addr": request.META.get('REMOTE_ADDR'),
            "request.content_length": request.META.get('CONTENT_LENGTH', 0),
            "request.user_agent": request.META.get('HTTP_USER_AGENT'),
            "request.scheme": request.scheme,
            "request.secure": request.is_secure(),
            "request.query": request.GET.dict(),
            "request.xhr": request.headers.get('x-requested-with') == 'XMLHttpRequest',
        }

    def get_context_from_response(self, request, response):
        return {
            "response.status_code": response.status_code,
        }

    def create_http_event(self, request):
        # if beeline has not been initialised, just execute request
        if not beeline.get_beeline():
            return self.get_response(request)

        # Code to be executed for each request before
        # the view (and later middleware) are called.
        dr = DjangoRequest(request)

        request_context = self.get_context_from_request(request)
        root_span = beeline.propagate_and_start_trace(request_context, dr)

        response = self.get_response(request)

        # Code to be executed for each request/response after
        # the view is called.
        response_context = self.get_context_from_response(request, response)
        beeline.add_context(response_context)

        # Streaming responses return immediately, but iterate over
        # their `streaming_content` until it's empty; only close the
        # trace then, not now.
        def wrap_streaming_content(content):
            for chunk in content:
                yield chunk
            beeline.finish_trace(root_span)

        if response.streaming:
            response.streaming_content = wrap_streaming_content(
                response.streaming_content
            )
        else:
            beeline.finish_trace(root_span)

        return response

    def process_exception(self, request, exception):
        if beeline.get_beeline():
            beeline.add_context_field(
                "request.error_detail", beeline.internal.stringify_exception(exception))

    def process_view(self, request, view_func, view_args, view_kwargs):
        if beeline.get_beeline():
            try:
                beeline.add_context_field("django.view_func", view_func.__name__)
            except AttributeError:
                pass

            try:
                beeline.add_context_field("request.route", request.resolver_match.route)
            except AttributeError:
                pass


class HoneyMiddlewareHttp(HoneyMiddlewareBase):
    pass


class HoneyMiddleware(HoneyMiddlewareBase):
    def __call__(self, request):
        try:
            db_wrapper = HoneyDBWrapper()
            # db instrumentation is only present in Django > 2.0
            with contextlib.ExitStack() as stack:
                for connection in connections.all():
                    stack.enter_context(connection.execute_wrapper(db_wrapper))
                response = self.create_http_event(request)

                # Is the response is streaming, then _this_ context
                # will exit now, but we want to set up the same sort
                # context stack when that streaming content gets
                # iterated over.
                def wrap_streaming_content(content):
                    with contextlib.ExitStack() as stack:
                        for connection in connections.all():
                            stack.enter_context(connection.execute_wrapper(db_wrapper))
                        for chunk in content:
                            yield chunk
                if response.streaming:
                    response.streaming_content = wrap_streaming_content(
                        response.streaming_content
                    )
        except AttributeError:
            response = self.create_http_event(request)

        return response


class HoneyMiddlewareWithPOST(HoneyMiddleware):
    ''' HoneyMiddlewareWithPOST is a subclass of HoneyMiddleware. The only difference is that
    the `request.post` field is instrumented. This was removed from the base implementation in 2.8.0
    due to conflicts with other middleware. See https://github.com/honeycombio/beeline-python/issues/74.'''

    def get_context_from_request(self, request):
        trace_name = "django_http_%s" % request.method.lower()
        return {
            "name": trace_name,
            "type": "http_server",
            "request.host": request.get_host(),
            "request.method": request.method,
            "request.path": request.path,
            "request.remote_addr": request.META.get('REMOTE_ADDR'),
            "request.content_length": request.META.get('CONTENT_LENGTH', 0),
            "request.user_agent": request.META.get('HTTP_USER_AGENT'),
            "request.scheme": request.scheme,
            "request.secure": request.is_secure(),
            "request.query": request.GET.dict(),
            "request.xhr": request.headers.get('x-requested-with') == 'XMLHttpRequest',
            "request.post": request.POST.dict(),
        }



import unittest
import beeline.propagation

header_value = '1;trace_id=bloop,parent_id=scoop,context=e30K'


class TestDictRequest(unittest.TestCase):
    def test_headers(self):
        '''Test that we correctly deal with case sensitivity'''
        request = beeline.propagation.DictRequest({
            # case shouldn't matter
            'MixedCaseHeader': "value",
            'UPPERCASEHEADER': "value"
        },
            {})
        value = request.header('mixedcaseheader')
        self.assertEqual(value, "value")
        value = request.header('upperCaseHeader')
        self.assertEqual(value, "value")

    def test_request_props(self):
        '''Test we correctly return request props'''
        request = beeline.propagation.DictRequest({
            # case shouldn't matter
            'MixedCaseHeader': "value",
            'UPPERCASEHEADER': "value"
        },
            {
                'method': "GET",
                'scheme': "http",
                'host': "api.honeycomb.io",
                'path': "/1/event",
                'query': "key=value"
        })

        self.assertEqual(request.method(), "GET")
        self.assertEqual(request.scheme(), "http")
        self.assertEqual(request.host(), "api.honeycomb.io")
        self.assertEqual(request.path(), "/1/event")
        self.assertEqual(request.query(), "key=value")



import base64
import copy
import datetime
import functools
import hashlib
import json
import math
import random
import struct
import sys
import threading
import traceback
import inspect
from collections import defaultdict

from contextlib import contextmanager

from beeline.internal import log, stringify_exception

import beeline.propagation
import beeline.propagation.default
import beeline.propagation.honeycomb

MAX_INT32 = math.pow(2, 32) - 1
SPAN_ID_BYTES = 8
TRACE_ID_BYTES = 16


class Trace(object):
    '''Object encapsulating all state of an ongoing trace.'''

    def __init__(self, trace_id, dataset=None):
        self.id = trace_id
        self.dataset = dataset
        self.stack = []
        self.fields = {}
        self.rollup_fields = defaultdict(float)

    def copy(self):
        '''Copy the trace state for use in another thread or context.'''
        result = Trace(self.id)
        result.stack = copy.copy(self.stack)
        result.fields = copy.copy(self.fields)
        return result


class Tracer(object):
    def __init__(self, client):
        self._client = client

        self.presend_hook = None
        self.sampler_hook = None
        self.http_trace_parser_hook = beeline.propagation.default.http_trace_parser_hook
        self.http_trace_propagation_hook = beeline.propagation.default.http_trace_propagation_hook

    @contextmanager
    def __call__(self, name, trace_id=None, parent_id=None):
        try:
            span = None
            if self.get_active_trace_id() and trace_id is None:
                span = self.start_span(
                    context={'name': name}, parent_id=parent_id)
                if span:
                    log('tracer context manager started new span, id = %s',
                        span.id)
            else:
                span = self.start_trace(
                    context={'name': name}, trace_id=trace_id, parent_span_id=parent_id)
                if span:
                    log('tracer context manager started new trace, id = %s',
                        span.trace_id)
            yield span
        except Exception as e:
            if span:
                span.add_context({
                    "app.exception_type": str(type(e)),
                    "app.exception_string": stringify_exception(e),
                    "app.exception_stacktrace": traceback.format_exc(),
                })
            raise
        finally:
            if span:
                if span.is_root():
                    log('tracer context manager ending trace, id = %s',
                        span.trace_id)
                    self.finish_trace(span)
                else:
                    log('tracer context manager ending span, id = %s',
                        span.id)
                    self.finish_span(span)
            else:
                log('tracer context manager span for %s was unexpectedly None', name)

    def start_trace(self, context=None, trace_id=None, parent_span_id=None, dataset=None):
        if trace_id:
            if self._trace:
                log('warning: start_trace got explicit trace_id but we are already in a trace. '
                    'starting new trace with id = %s', trace_id)
            self._trace = Trace(trace_id, dataset)
        else:
            self._trace = Trace(generate_trace_id(), dataset)

        # start the root span
        return self.start_span(context=context, parent_id=parent_span_id)

    def start_span(self, context=None, parent_id=None):
        if not self._trace:
            log('start_span called but no trace is active')
            return None

        span_id = generate_span_id()
        if parent_id:
            parent_span_id = parent_id
        else:
            parent_span_id = self._trace.stack[-1].id if self._trace.stack else None
        ev = self._client.new_event(data=self._trace.fields)
        if context:
            ev.add(data=context)

        ev.add(data={
            'trace.trace_id': self._trace.id,
            'trace.parent_id': parent_span_id,
            'trace.span_id': span_id,
        })
        is_root = len(self._trace.stack) == 0
        span = Span(trace_id=self._trace.id, parent_id=parent_span_id,
                    id=span_id, event=ev, is_root=is_root)
        self._trace.stack.append(span)

        return span

    def finish_span(self, span):
        # avoid exception if called with None
        if span is None:
            return

        # send the span's event. Even if the stack is in an unhealthy state,
        # it's probably better to send event data than not
        if span.event:
            if self._trace:
                if self._trace.dataset:
                    span.event.dataset = self._trace.dataset

                # add the trace's rollup fields to the root span
                if span.is_root():
                    for k, v in self._trace.rollup_fields.items():
                        span.event.add_field(k, v)

                for k, v in span.rollup_fields.items():
                    span.event.add_field(k, v)

                # propagate trace fields that may have been added in later spans
                for k, v in self._trace.fields.items():
                    # don't overwrite existing values because they may be different
                    if k not in span.event.fields():
                        span.event.add_field(k, v)

            duration = datetime.datetime.now() - span.event.start_time
            duration_ms = duration.total_seconds() * 1000.0
            span.event.add_field('duration_ms', duration_ms)

            self._run_hooks_and_send(span)
        else:
            log('warning: span has no event, was it initialized correctly?')

        if not self._trace:
            log('warning: span finished without an active trace')
            return

        if span.trace_id != self._trace.id:
            log('warning: finished span called for span in inactive trace. '
                'current trace_id = %s, span trace_id = %s', self._trace.id, span.trace_id)
            return

        if not self._trace.stack:
            log('warning: finish span called but stack is empty')
            return

        if self._trace.stack[-1].id != span.id:
            log('warning: finished span is not the currently active span')
            return

        self._trace.stack.pop()

    def finish_trace(self, span):
        self.finish_span(span)
        self._trace = None

    def parse_http_trace(self, request):
        if not self.http_trace_parser_hook:
            return None
        return self.http_trace_parser_hook(request)

    def propagate_and_start_trace(self, context, request):
        err = None
        propagation_context = None
        try:
            propagation_context = self.parse_http_trace(request)
        except Exception:
            err = sys.exc_info()[0]
            log('error: http_trace_parser_hook returned exception: %s',
                sys.exc_info()[0])

        if propagation_context:
            return self.start_trace(context=context, trace_id=propagation_context.trace_id,
                                    parent_span_id=propagation_context.parent_id,
                                    dataset=propagation_context.dataset)
            for k, v in propagation_context.trace_fields:
                self.add_trace_field(k, v)
        else:
            # Initialize a new trace from scratch
            if err is not None:
                context['parser_hook_error'] = repr(err)
            return self.start_trace(context, trace_id=None, parent_span_id=None)
        pass

    def get_propagation_context(self):
        if not self._trace:
            return None

        return beeline.propagation.PropagationContext(
            self.get_active_trace_id(),
            self.get_active_span().id,
            self._trace.fields)

    def get_active_trace_id(self):
        if self._trace:
            return self._trace.id
        return None

    def get_active_span(self):
        if self._trace and self._trace.stack:
            return self._trace.stack[-1]
        return None

    def add_context_field(self, name, value):
        span = self.get_active_span()
        if span:
            span.add_context_field(name=name, value=value)

    def add_context(self, data):
        span = self.get_active_span()
        if span:
            span.add_context(data=data)

    def remove_context_field(self, name):
        span = self.get_active_span()
        if span:
            span.remove_context_field(name=name)

    def add_rollup_field(self, name, value):
        value = float(value)

        span = self.get_active_span()
        if span:
            span.rollup_fields[name] += value

        if not self._trace:
            log('warning: adding rollup field without an active trace')
            return

        self._trace.rollup_fields["rollup.%s" % name] += value

    def add_trace_field(self, name, value):
        # prefix with app to avoid key conflicts
        # add the app prefix if it's missing

        if (type(name) == str and not name.startswith("app.")) or type(name) != str:
            key = "app.%s" % name
        else:
            key = name

        # also add to current span
        self.add_context_field(key, value)

        if not self._trace:
            log('warning: adding trace field without an active trace')
            return
        self._trace.fields[key] = value

    def remove_trace_field(self, name):
        key = "app.%s" % name
        self.remove_context_field(key)
        if not self._trace:
            log('warning: removing trace field without an active trace')
            return
        self._trace.fields.pop(key)

    def marshal_trace_context(self):
        if not self._trace:
            log('warning: marshal_trace_context called, but no active trace')
            return

        return marshal_trace_context(
            self._trace.id,
            self._trace.stack[-1].id,
            self._trace.fields
        )

    def register_hooks(self, presend=None, sampler=None, http_trace_parser=None, http_trace_propagation=None):
        self.presend_hook = presend
        self.sampler_hook = sampler
        self.http_trace_parser_hook = http_trace_parser
        self.http_trace_propagation_hook = http_trace_propagation

    def _run_hooks_and_send(self, span):
        ''' internal - run any defined hooks on the event and send

        kind of hacky: we fetch the hooks from the beeline, but they are only
        used here. Pass them to the tracer implementation?
        '''
        presampled = False
        if self.sampler_hook:
            log("executing sampler hook on event ev = %s", span.event.fields())
            keep, new_rate = self.sampler_hook(span.event.fields())
            if not keep:
                log("skipping event due to sampler hook sampling ev = %s",
                    span.event.fields())
                return
            span.event.sample_rate = new_rate
            presampled = True

        if self.presend_hook:
            log("executing presend hook on event ev = %s", span.event.fields())
            self.presend_hook(span.event.fields())

        if presampled:
            log("enqueuing presampled event ev = %s", span.event.fields())
            span.event.send_presampled()
        elif _should_sample(span.trace_id, span.event.sample_rate):
            # if our sampler hook wasn't used, use deterministic sampling
            span.event.send_presampled()


class SynchronousTracer(Tracer):
    def __init__(self, client):
        super().__init__(client)
        self._state = threading.local()

    @property
    def _trace(self):
        return getattr(self._state, 'trace', None)

    @_trace.setter
    def _trace(self, new_trace):
        self._state.trace = new_trace


class Span(object):
    ''' Span represents an active span. Should not be initialized directly, but
    through a Tracer object's `start_span` method. '''

    def __init__(self, trace_id, parent_id, id, event, is_root=False):
        self.trace_id = trace_id
        self.parent_id = parent_id
        self.id = id
        self.event = event
        self.event.start_time = datetime.datetime.now()
        self.rollup_fields = defaultdict(float)
        self._is_root = is_root

    def add_context_field(self, name, value):
        self.event.add_field(name, value)

    def add_context(self, data):
        self.event.add(data)

    def remove_context_field(self, name):
        if name in self.event.fields():
            del self.event.fields()[name]

    def is_root(self):
        return self._is_root


def _should_sample(trace_id, sample_rate):
    # Always return True if sample rate is 1
    if sample_rate == 1:
        return True

    sample_upper_bound = MAX_INT32 / sample_rate
    # compute a sha1
    sha1 = hashlib.sha1()
    sha1.update(trace_id.encode('utf-8'))
    # convert first 4 digits to int
    value, = struct.unpack('>I', sha1.digest()[:4])
    if value < sample_upper_bound:
        return True
    return False


def marshal_trace_context(trace_id, parent_id, context):
    """Deprecated: Use beeline.propagation.honeycomb.marshal_trace_context instead"""
    version = 1
    trace_fields = base64.b64encode(json.dumps(context).encode()).decode()
    trace_context = "{};trace_id={},parent_id={},context={}".format(
        version, trace_id, parent_id, trace_fields
    )

    return trace_context


def unmarshal_trace_context(trace_header):
    """
    Deprecated: Use beeline.propagation.honeycomb.unmarshal_trace_context instead
    """
    return beeline.propagation.honeycomb.unmarshal_propagation_context(trace_header)


def traced_impl(tracer_fn, name, trace_id, parent_id):
    """Implementation of the traced decorator without async support."""
    def wrapped(fn):
        if inspect.isgeneratorfunction(fn):
            @functools.wraps(fn)
            def inner(*args, **kwargs):
                inner_generator = fn(*args, **kwargs)
                with tracer_fn(name=name, trace_id=trace_id, parent_id=parent_id):
                    for value in inner_generator:
                        yield value
            return inner
        else:
            @functools.wraps(fn)
            def inner(*args, **kwargs):
                with tracer_fn(name=name, trace_id=trace_id, parent_id=parent_id):
                    return fn(*args, **kwargs)
            return inner
    return wrapped


# Use system random instead of default psuedorandom generator
system_random = random.SystemRandom()


def generate_span_id():
    """Generate span ID compatible with w3c tracing spec."""
    format_str = "{{:0{:d}x}}".format(SPAN_ID_BYTES*2)
    return format_str.format(system_random.getrandbits(SPAN_ID_BYTES*8))


def generate_trace_id():
    """Generate trace ID compatible with w3c tracing spec."""
    format_str = "{{:0{:d}x}}".format(TRACE_ID_BYTES*2)
    return format_str.format(system_random.getrandbits(TRACE_ID_BYTES*8))



from __future__ import absolute_import
import unittest
import unittest.loader
import sys


def get_test_suite():
    """Return the set of tests suitable for the current Python version"""
    # FIXME: Automatically discover tests and strip out async

    test_suite = unittest.defaultTestLoader.discover(".")

    filtered_test_suite = unittest.TestSuite()
    for test_group in test_suite:
        ts = unittest.TestSuite()
        for inner_test_group in test_group:
            # Skip the async tests which fail to import on old versions of python
            if inner_test_group.__class__.__name__ == "ModuleImportFailure" and inner_test_group._testMethodName == "beeline.test_async":
                print(
                    "Skipping beeline.test_async module tests due to old Python version")
            else:
                ts.addTest(inner_test_group)
        filtered_test_suite.addTest(ts)

    return filtered_test_suite


def run_tests():
    runner = unittest.TextTestRunner()
    return runner.run(get_test_suite())


if __name__ == "__main__":
    result = run_tests()
    sys.exit(not result.wasSuccessful())



import beeline

# these are mostly convenience methods for safely calling beeline methods
# even if the beeline hasn't been initialized


def send_event():
    bl = beeline.get_beeline()
    if bl:
        return bl.send_event()


def send_all():
    bl = beeline.get_beeline()
    if bl:
        return bl.send_all()


def log(msg, *args, **kwargs):
    bl = beeline.get_beeline()
    if bl:
        bl.log(msg, *args, **kwargs)


def stringify_exception(e):
    try:
        return str(e)
    except UnicodeEncodeError:
        try:
            return u"{}".format(e)
        except Exception:
            return "unable to decode exception"



"""Asynchronous tracer implementation.

This requires Python 3.7, because it uses the contextvars module.

"""
import asyncio
import contextvars  # pylint: disable=import-error
import functools
import inspect

from beeline.trace import Tracer

current_trace_var = contextvars.ContextVar("current_trace")


def create_task_factory(parent_factory):
    """Create a task factory that makes a copy of the current trace.

    New tasks have their own context variables, but the current_trace
    context variable still refers to the same Trace object as the one
    in the parent task. This task factory replaces the Trace object
    with a copy of itself.

    """
    def task_factory_impl(loop, coro):
        async def wrapper():
            current_trace = current_trace_var.get(None)
            if current_trace is not None:
                current_trace_var.set(current_trace.copy())
            return await coro

        if parent_factory is None:
            task = asyncio.tasks.Task(wrapper(), loop=loop)
        else:
            task = parent_factory(wrapper())

        return task

    task_factory_impl.__trace_task_factory__ = True
    return task_factory_impl


class AsyncioTracer(Tracer):
    def __init__(self, client):
        """Initialize, and ensure that our task factory is set up."""
        super().__init__(client)

        loop = asyncio.get_running_loop()  # pylint: disable=no-member

        task_factory = loop.get_task_factory()
        if task_factory is None or not task_factory.__trace_task_factory__:
            new_task_factory = create_task_factory(task_factory)
            loop.set_task_factory(new_task_factory)

    @property
    def _trace(self):
        return current_trace_var.get(None)

    @_trace.setter
    def _trace(self, new_trace):
        current_trace_var.set(new_trace)


def traced_impl(tracer_fn, name, trace_id, parent_id):
    """Implementation of the traced decorator including async support.

    The async version needs to be different, because the trace should
    cover the execution of the whole decorated function. If using the
    synchronous version, the trace would only cover the time it takes
    to return the coroutine object.

    """
    def wrapped(fn):
        if asyncio.iscoroutinefunction(fn):
            @functools.wraps(fn)
            async def async_inner(*args, **kwargs):
                with tracer_fn(name=name, trace_id=trace_id, parent_id=parent_id):
                    return await fn(*args, **kwargs)

            return async_inner
        elif inspect.isgeneratorfunction(fn):
            @functools.wraps(fn)
            def inner(*args, **kwargs):
                inner_generator = fn(*args, **kwargs)
                with tracer_fn(name=name, trace_id=trace_id, parent_id=parent_id):
                    yield from inner_generator

            return inner
        else:
            @functools.wraps(fn)
            def inner(*args, **kwargs):
                with tracer_fn(name=name, trace_id=trace_id, parent_id=parent_id):
                    return fn(*args, **kwargs)

            return inner

    return wrapped


def untraced(fn):
    """Async function decorator detaching from any ongoing trace.

    This decorator is necessary for starting independent async tasks
    from within a trace, since async tasks inherit trace state by
    default.

    """

    # Both synchronous and asynchronous functions may create tasks.
    if asyncio.iscoroutinefunction(fn):
        @functools.wraps(fn)
        async def wrapped(*args, **kwargs):
            try:
                token = None
                current_trace = current_trace_var.get(None)
                if current_trace is not None:
                    token = current_trace_var.set(None)

                return await fn(*args, **kwargs)
            finally:
                if token is not None:
                    current_trace_var.reset(token)

        return wrapped

    else:
        @functools.wraps(fn)
        def wrapped(*args, **kwargs):
            try:
                token = None
                current_trace = current_trace_var.get(None)
                if current_trace is not None:
                    token = current_trace_var.set(None)

                return fn(*args, **kwargs)
            finally:
                if token is not None:
                    current_trace_var.reset(token)

        return wrapped
